import logging
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from functools import partial

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import ResourceDoesNotExist, NotFound
from databricks.sdk.service.catalog import (
    ExternalLocationInfo,
    SchemaInfo,
    TableInfo,
    Privilege,
    PermissionsChange,
    SecurableType,
)
from databricks.sdk.service.compute import ClusterSource, DataSecurityMode

from databricks.labs.ucx.assessment.aws import AWSRoleAction
from databricks.labs.ucx.assessment.azure import (
    AzureServicePrincipalCrawler,
    AzureServicePrincipalInfo,
)
from databricks.labs.ucx.aws.access import AWSResourcePermissions
from databricks.labs.ucx.azure.access import (
    AzureResourcePermissions,
    StoragePermissionMapping,
)
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.locations import (
    ExternalLocations,
    Mount,
    Mounts,
)
from databricks.labs.ucx.hive_metastore.tables import Table, TablesCrawler
from databricks.labs.ucx.hive_metastore.udfs import UdfsCrawler

logger = logging.getLogger(__name__)


@dataclass
class LocationACL:
    location_name: str
    principal: str


@dataclass(frozen=True)
class Grant:
    principal: str
    action_type: str
    catalog: str | None = None
    database: str | None = None
    table: str | None = None
    view: str | None = None
    udf: str | None = None
    any_file: bool = False
    anonymous_function: bool = False

    @staticmethod
    def type_and_key(
        *,
        catalog: str | None = None,
        database: str | None = None,
        table: str | None = None,
        view: str | None = None,
        udf: str | None = None,
        any_file: bool = False,
        anonymous_function: bool = False,
    ) -> tuple[str, str]:
        if table is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            database = "default" if database is None else database
            return "TABLE", f"{catalog}.{database}.{table}"
        if view is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            database = "default" if database is None else database
            return "VIEW", f"{catalog}.{database}.{view}"
        if udf is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            database = "default" if database is None else database
            return "FUNCTION", f"{catalog}.{database}.{udf}"
        if database is not None:
            catalog = "hive_metastore" if catalog is None else catalog
            return "DATABASE", f"{catalog}.{database}"
        if any_file:
            return "ANY FILE", ""
        if anonymous_function:
            return "ANONYMOUS FUNCTION", ""
        # Must come last, as it has the lowest priority here but is a required parameter
        if catalog is not None:
            return "CATALOG", catalog
        msg = (
            f"invalid grant keys: catalog={catalog}, database={database}, view={view}, udf={udf}"
            f"any_file={any_file}, anonymous_function={anonymous_function}"
        )
        raise ValueError(msg)

    @property
    def object_key(self) -> str:
        _, key = self.this_type_and_key()
        return key.lower()

    def this_type_and_key(self):
        return self.type_and_key(
            catalog=self.catalog,
            database=self.database,
            table=self.table,
            view=self.view,
            udf=self.udf,
            any_file=self.any_file,
            anonymous_function=self.anonymous_function,
        )

    def hive_grant_sql(self) -> list[str]:
        object_type, object_key = self.this_type_and_key()
        # See https://docs.databricks.com/en/sql/language-manual/security-grant.html
        statements = []
        actions = self.action_type.split(", ")
        deny_actions = [action for action in actions if "DENIED" in action]
        grant_actions = [action for action in actions if "DENIED" not in action]
        if "OWN" in grant_actions:
            grant_actions.remove("OWN")
            statements.append(self._set_owner_sql(object_type, object_key))
        if grant_actions:
            statements.append(self._apply_grant_sql(", ".join(grant_actions), object_type, object_key))
        if deny_actions:
            statements.append(self._apply_grant_sql(", ".join(deny_actions), object_type, object_key))
        return statements

    def hive_revoke_sql(self) -> str:
        object_type, object_key = self.this_type_and_key()
        return f"REVOKE {self.action_type} ON {object_type} {escape_sql_identifier(object_key)} FROM `{self.principal}`"

    def _set_owner_sql(self, object_type, object_key):
        return f"ALTER {object_type} {escape_sql_identifier(object_key)} OWNER TO `{self.principal}`"

    def _apply_grant_sql(self, action_type, object_type, object_key):
        if "DENIED" in action_type:
            action_type = action_type.replace("DENIED_", "")
            # need to wrap the action type in backticks to avoid syntax errors with DENY SELECT
            return f"DENY `{action_type}` ON {object_type} {escape_sql_identifier(object_key)} TO `{self.principal}`"
        return f"GRANT {action_type} ON {object_type} {escape_sql_identifier(object_key)} TO `{self.principal}`"

    def _uc_action(self, action_type):
        def inner(object_type, object_key):
            return self._apply_grant_sql(action_type, object_type, object_key)

        return inner

    def uc_grant_sql(self, object_type: str | None = None, object_key: str | None = None) -> str | None:
        """Get SQL translated SQL statement for granting similar permissions in UC.

        If there's no UC equivalent, returns None. This can also be the case for missing mapping.
        """

        # TODO: verify and complete the mapping
        # See: https://docs.databricks.com/sql/language-manual/sql-ref-privileges-hms.html
        # See: https://docs.databricks.com/data-governance/unity-catalog/manage-privileges/ownership.html
        # See: https://docs.databricks.com/data-governance/unity-catalog/manage-privileges/privileges.html
        if object_type is None:
            object_type, object_key = self.this_type_and_key()
        hive_to_uc = {
            ("FUNCTION", "SELECT"): self._uc_action("EXECUTE"),
            ("TABLE", "SELECT"): self._uc_action("SELECT"),
            ("TABLE", "MODIFY"): self._uc_action("MODIFY"),
            ("TABLE", "ALL PRIVILEGES"): self._uc_action("ALL PRIVILEGES"),
            ("TABLE", "OWN"): self._set_owner_sql,
            ("VIEW", "SELECT"): self._uc_action("SELECT"),
            ("VIEW", "OWN"): self._set_owner_sql,
            ("DATABASE", "USAGE"): self._uc_action("USE SCHEMA"),
            ("DATABASE", "CREATE"): self._uc_action("CREATE TABLE"),
            ("DATABASE", "CREATE_NAMED_FUNCTION"): self._uc_action("CREATE FUNCTION"),
            ("DATABASE", "SELECT"): self._uc_action("SELECT"),
            ("DATABASE", "MODIFY"): self._uc_action("MODIFY"),
            ("DATABASE", "OWN"): self._set_owner_sql,
            ("CATALOG", "OWN"): self._set_owner_sql,
            ("CATALOG", "USAGE"): self._uc_action("USE CATALOG"),
        }
        make_query = hive_to_uc.get((object_type, self.action_type), None)
        if make_query is None:
            # unknown mapping or ignore
            return None
        return make_query(object_type, object_key)


class GrantsCrawler(CrawlerBase[Grant]):
    def __init__(self, tc: TablesCrawler, udf: UdfsCrawler, include_databases: list[str] | None = None):
        assert tc._backend == udf._backend
        assert tc._catalog == udf._catalog
        assert tc._schema == udf._schema
        super().__init__(tc._backend, tc._catalog, tc._schema, "grants", Grant)
        self._tc = tc
        self._udf = udf
        self._include_databases = include_databases

    def snapshot(self) -> Iterable[Grant]:
        return self._snapshot(partial(self._try_load), partial(self._crawl))

    def _try_load(self):
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Grant(*row)

    def _crawl(self) -> Iterable[Grant]:
        """
        Crawls and lists grants for all databases, tables,  views, udfs, any file
        and anonymous function within hive_metastore.

        Returns:
            list[Grant]: A list of Grant objects representing the listed grants.

        Behavior:
        - Constructs a list of tasks to fetch grants using the `_grants` method, including both database-wide and
          table/view-specific grants.
        - Iterates through tables in the specified database using the `_tc.snapshot` method.
        - For each table, adds tasks to fetch grants for the table or its view, depending on the kind of the table.
        - Iterates through udfs in the specified database using the `_udf.snapshot` method.
        - For each udf, adds tasks to fetch grants for the udf.
        - Executes the tasks concurrently using Threads.gather.
        - Flattens the list of retrieved grant lists into a single list of Grant objects.

        Note:
        - The method assumes that the `_grants` method fetches grants based on the provided parameters (catalog,
          database, table, view, udfs, any file, anonymous function).

        Returns:
        list[Grant]: A list of Grant objects representing the grants found in hive_metastore.
        """
        catalog = "hive_metastore"
        tasks = [partial(self.grants, catalog=catalog)]
        # Scanning ANY FILE and ANONYMOUS FUNCTION grants
        tasks.append(partial(self.grants, catalog=catalog, any_file=True))
        tasks.append(partial(self.grants, catalog=catalog, anonymous_function=True))
        if not self._include_databases:
            # scan all databases, even empty ones
            for row in self._fetch(f"SHOW DATABASES FROM {escape_sql_identifier(catalog)}"):
                tasks.append(partial(self.grants, catalog=catalog, database=row.databaseName))
        else:
            for database in self._include_databases:
                tasks.append(partial(self.grants, catalog=catalog, database=database))
        for table in self._tc.snapshot():
            fn = partial(self.grants, catalog=catalog, database=table.database)
            # Views are treated as a type of table and enumerated by the table crawler.
            if table.view_text is None:
                task = partial(fn, table=table.name)
            else:
                task = partial(fn, view=table.name)
            tasks.append(task)
        for udf in self._udf.snapshot():
            fn = partial(self.grants, catalog=catalog, database=udf.database)
            tasks.append(partial(fn, udf=udf.name))
        catalog_grants, errors = Threads.gather(f"listing grants for {catalog}", tasks)
        if len(errors) > 0:
            raise ManyError(errors)
        return [grant for grants in catalog_grants for grant in grants]

    def for_table_info(self, table: TableInfo):
        # TODO: it does not work yet for views
        principal_permissions = defaultdict(set)
        for grant in self.grants(catalog=table.catalog_name, database=table.schema_name, table=table.name):
            principal_permissions[grant.principal].add(grant.action_type)
        return principal_permissions

    def for_schema_info(self, schema: SchemaInfo):
        principal_permissions = defaultdict(set)
        for grant in self.grants(catalog=schema.catalog_name, database=schema.name):
            principal_permissions[grant.principal].add(grant.action_type)
        return principal_permissions

    # The reported ObjectType for grants doesn't necessary match the type of grants we asked for. This maps
    # those that aren't themselves.
    _grants_reported_as = {
        # SHOW type: RESULT type
        "SCHEMA": "DATABASE",
        "CATALOG": "CATALOG$",
        "VIEW": "TABLE",
        "ANY FILE": "ANY_FILE",
        "ANONYMOUS FUNCTION": "ANONYMOUS_FUNCTION",
    }

    def grants(
        self,
        *,
        catalog: str | None = None,
        database: str | None = None,
        table: str | None = None,
        view: str | None = None,
        udf: str | None = None,
        any_file: bool = False,
        anonymous_function: bool = False,
    ) -> list[Grant]:
        """
        Fetches and yields grant information for the specified database objects.

        Keyword Args:
            catalog (str): The catalog name (optional).
            database (str | None): The database name (optional).
            table (str | None): The table name (optional).
            view (str | None): The view name (optional).
            udf (str | None): The udf name (optional).
            any_file (bool): Whether to include any file grants (optional).
            anonymous_function (bool): Whether to include anonymous function grants (optional).

        Behavior:
        - Normalizes the provided parameters and constructs an object type and key using
          the `Grant.type_and_key` method.
        - Iterates through rows fetched using the `_fetch` method by executing a SQL query
          to retrieve grant information.
        - For each fetched row, extracts the principal, action type, and object type.
        - Normalizes the object type and filters grants based on the provided object type
          and the fetched object type.
        - Yields a Grant object representing the fetched grant information.

        Note:
        - The method fetches and yields grants based on the provided parameters and
          the available grant information in the database.

        Returns:
        Iterator[Grant]: An iterator of Grant objects representing the fetched grant information.
        """
        on_type, key = Grant.type_and_key(
            catalog=self._try_valid(catalog),
            database=self._try_valid(database),
            table=self._try_valid(table),
            view=self._try_valid(view),
            udf=self._try_valid(udf),
            any_file=any_file,
            anonymous_function=anonymous_function,
        )
        expected_grant_object_type = self._grants_reported_as.get(on_type, on_type)
        grants = []
        try:
            for row in self._fetch(f"SHOW GRANTS ON {on_type} {escape_sql_identifier(key)}"):
                (principal, action_type, object_type, _) = row
                # This seems to be here to help unit tests that don't sufficiently narrow the target of mocked queries. -ajs
                if object_type != expected_grant_object_type:
                    continue
                # we have to return concrete list, as with yield we're executing
                # everything on the main thread.
                grant = Grant(
                    principal=principal,
                    action_type=action_type,
                    table=table,
                    view=view,
                    udf=udf,
                    database=database,
                    catalog=catalog,
                    any_file=any_file,
                    anonymous_function=anonymous_function,
                )
                grants.append(grant)
            return grants
        except NotFound:
            # This make the integration test more robust as many test schemas are being created and deleted quickly.
            logger.warning(f"Schema {catalog}.{database} no longer existed")
            return []
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(f"Couldn't fetch grants for object {on_type} {key}: {e}")
            return []


class AwsACL:
    def __init__(
        self,
        ws: WorkspaceClient,
        backend: SqlBackend,
        installation: Installation,
    ):
        self._backend = backend
        self._ws = ws
        self._installation = installation

    def _get_cluster_to_instance_profile_mapping(self) -> dict[str, str]:
        # this function gives a mapping between an interactive cluster and the instance profile used by it
        # either directly or through a cluster policy.
        cluster_instance_profiles = {}
        for cluster in self._ws.clusters.list():
            if (
                cluster.cluster_id is None
                or cluster.cluster_source == ClusterSource.JOB
                or (cluster.data_security_mode not in [DataSecurityMode.LEGACY_SINGLE_USER, DataSecurityMode.NONE])
            ):
                continue
            if cluster.aws_attributes is None:
                continue
            if cluster.aws_attributes.instance_profile_arn is not None:
                role_name = cluster.aws_attributes.instance_profile_arn
                assert role_name is not None
                cluster_instance_profiles[cluster.cluster_id] = role_name
                continue

        return cluster_instance_profiles

    def get_eligible_locations_principals(self) -> dict[str, dict]:
        cluster_locations = {}
        eligible_locations = {}
        cluster_instance_profiles = self._get_cluster_to_instance_profile_mapping()
        if len(cluster_instance_profiles) == 0:
            # if there are no interactive clusters , then return empty grants
            logger.info("No interactive cluster found with instance profiles configured")
            return {}
        external_locations = list(self._ws.external_locations.list())
        if len(external_locations) == 0:
            # if there are no external locations, then throw an error to run migrate_locations cli command
            msg = (
                "No external location found, If hive metastore tables are created in external storage, "
                "ensure migrate-locations cli cmd is run to create the required locations."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        permission_mappings = self._installation.load(
            list[AWSRoleAction],
            filename=AWSResourcePermissions.INSTANCE_PROFILES_FILE_NAME,
        )
        if len(permission_mappings) == 0:
            # if permission mapping is empty, raise an error to run principal_prefix cmd
            msg = (
                "No instance profile permission file found. Please ensure principal-prefix-access cli "
                "cmd is run to create the instance profile permission file."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        for cluster_id, role_name in cluster_instance_profiles.items():
            eligible_locations.update(self._get_external_locations(role_name, external_locations, permission_mappings))
            if len(eligible_locations) == 0:
                continue
            cluster_locations[cluster_id] = eligible_locations
        return cluster_locations

    @staticmethod
    def _get_external_locations(
        role_name: str,
        external_locations: list[ExternalLocationInfo],
        permission_mappings: list[AWSRoleAction],
    ) -> dict[str, str]:
        matching_location = {}
        for location in external_locations:
            if location.url is None:
                continue
            for permission_mapping in permission_mappings:
                # check if resource_path contains "*"
                if permission_mapping.resource_path.find('*') > 1:
                    resource_url = permission_mapping.resource_path[: permission_mapping.resource_path.index('*') - 1]
                else:
                    resource_url = permission_mapping.resource_path
                if location.url.startswith(resource_url) and permission_mapping.role_arn == role_name:
                    matching_location[location.url] = permission_mapping.privilege
        return matching_location


class AzureACL:
    def __init__(
        self,
        ws: WorkspaceClient,
        backend: SqlBackend,
        spn_crawler: AzureServicePrincipalCrawler,
        installation: Installation,
    ):
        self._backend = backend
        self._ws = ws
        self._spn_crawler = spn_crawler
        self._installation = installation

    def get_eligible_locations_principals(self) -> dict[str, dict]:
        cluster_locations = {}
        eligible_locations = {}
        spn_cluster_mapping = self._spn_crawler.get_cluster_to_storage_mapping()
        if len(spn_cluster_mapping) == 0:
            # if there are no interactive clusters , then return empty grants
            logger.info("No interactive cluster found with spn configured")
            return {}
        external_locations = list(self._ws.external_locations.list())
        if len(external_locations) == 0:
            # if there are no external locations, then throw an error to run migrate_locations cli command
            msg = (
                "No external location found, If hive metastore tables are created in external storage, "
                "ensure migrate-locations cli cmd is run to create the required locations."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        permission_mappings = self._installation.load(
            list[StoragePermissionMapping],
            filename=AzureResourcePermissions.FILENAME,
        )
        if len(permission_mappings) == 0:
            # if permission mapping is empty, raise an error to run principal_prefix cmd
            msg = (
                "No storage permission file found. Please ensure principal-prefix-access cli "
                "cmd is run to create the access permission file."
            )
            logger.error(msg)
            raise ResourceDoesNotExist(msg) from None

        for cluster_spn in spn_cluster_mapping:
            for spn in cluster_spn.spn_info:
                eligible_locations.update(self._get_external_locations(spn, external_locations, permission_mappings))
            cluster_locations[cluster_spn.cluster_id] = eligible_locations
        return cluster_locations

    def _get_external_locations(
        self,
        spn: AzureServicePrincipalInfo,
        external_locations: list[ExternalLocationInfo],
        permission_mappings: list[StoragePermissionMapping],
    ) -> dict[str, str]:
        matching_location = {}
        for location in external_locations:
            if location.url is None:
                continue
            for permission_mapping in permission_mappings:
                prefix = permission_mapping.prefix
                if (
                    location.url.startswith(permission_mapping.prefix)
                    and permission_mapping.client_id == spn.application_id
                    and spn.storage_account is not None
                    # check for storage account name starting after @ in the prefix url
                    and prefix[prefix.index('@') + 1 :].startswith(spn.storage_account)
                ):
                    matching_location[location.url] = permission_mapping.privilege
        return matching_location


class PrincipalACL:
    def __init__(
        self,
        ws: WorkspaceClient,
        backend: SqlBackend,
        installation: Installation,
        tables_crawler: TablesCrawler,
        mounts_crawler: Mounts,
        cluster_locations: dict[str, dict],
    ):
        self._backend = backend
        self._ws = ws
        self._installation = installation
        self._tables_crawler = tables_crawler
        self._mounts_crawler = mounts_crawler
        self._cluster_locations = cluster_locations

    def get_interactive_cluster_grants(self) -> list[Grant]:
        tables = self._tables_crawler.snapshot()
        mounts = list(self._mounts_crawler.snapshot())
        grants: set[Grant] = set()

        for cluster_id, locations in self._cluster_locations.items():
            principals = self._get_cluster_principal_mapping(cluster_id)
            if len(principals) == 0:
                continue
            cluster_usage = self._get_grants(locations, principals, tables, mounts)
            grants.update(cluster_usage)
        return list(grants)

    def _get_privilege(self, table: Table, locations: dict[str, str], mounts: list[Mount]):
        if table.view_text is not None:
            # return nothing for view so that it goes to the separate view logic
            return None
        if table.location is None:
            return None
        if table.location.startswith('dbfs:/mnt') or table.location.startswith('/dbfs/mnt'):
            mount_location = ExternalLocations.resolve_mount(table.location, mounts)
            for loc, privilege in locations.items():
                if loc is not None and mount_location.startswith(loc):
                    return privilege
            return None
        if table.location.startswith('dbfs:/') or table.location.startswith('/dbfs/'):
            return "WRITE_FILES"

        for loc, privilege in locations.items():
            if loc is not None and table.location.startswith(loc):
                return privilege
        return None

    def _get_database_grants(self, tables: list[Table], principals: list[str]) -> list[Grant]:
        databases = {table.database for table in tables}
        return [
            Grant(principal, "USAGE", "hive_metastore", database) for database in databases for principal in principals
        ]

    def _get_grants(
        self, locations: dict[str, str], principals: list[str], tables: list[Table], mounts: list[Mount]
    ) -> list[Grant]:
        grants = []
        filtered_tables = []
        for table in tables:
            privilege = self._get_privilege(table, locations, mounts)
            if privilege == "READ_FILES":
                grants.extend(
                    [Grant(principal, "SELECT", table.catalog, table.database, table.name) for principal in principals]
                )
                filtered_tables.append(table)
                continue
            if privilege == "WRITE_FILES":
                grants.extend(
                    [
                        Grant(principal, "ALL PRIVILEGES", table.catalog, table.database, table.name)
                        for principal in principals
                    ]
                )
                filtered_tables.append(table)
                continue
            if table.view_text is not None:
                grants.extend(
                    [
                        Grant(principal, "ALL PRIVILEGES", table.catalog, table.database, view=table.name)
                        for principal in principals
                    ]
                )
                filtered_tables.append(table)

        database_grants = self._get_database_grants(filtered_tables, principals)

        grants.extend(database_grants)

        return grants

    def _get_cluster_principal_mapping(self, cluster_id: str) -> list[str]:
        # gets all the users,groups,spn which have access to the clusters and returns a dataclass of that mapping
        principal_list = []
        try:
            cluster_permission = self._ws.permissions.get("clusters", cluster_id)
        except ResourceDoesNotExist:
            return []
        if cluster_permission.access_control_list is None:
            return []
        for acl in cluster_permission.access_control_list:
            if acl.user_name is not None:
                principal_list.append(acl.user_name)
            if acl.group_name is not None:
                if acl.group_name == "admins":
                    continue
                principal_list.append(acl.group_name)
            if acl.service_principal_name is not None:
                principal_list.append(acl.service_principal_name)
        return principal_list

    def apply_location_acl(self):
        """
        Check the interactive cluster and the principals mapped to it
        identifies the spn or instance profile configured for the interactive cluster
        identifies any location the spn/instance profile have access to (read or write)
        applies create_external_table, create_external_volume and read_files permission for all location
        to the principal
        """
        logger.info(
            "Applying permission for external location (CREATE EXTERNAL TABLE, "
            "CREATE EXTERNAL VOLUME and READ_FILES for existing eligible interactive cluster users"
        )
        # get the eligible location mapped for each interactive cluster
        for cluster_id, locations in self._cluster_locations.items():
            # get interactive cluster users
            principals = self._get_cluster_principal_mapping(cluster_id)
            if len(principals) == 0:
                continue
            for location_url in locations.keys():
                # get the location name for the given url
                location_name = self._get_location_name(location_url)
                if location_name is None:
                    continue
                self._update_location_permissions(location_name, principals)
        logger.info("Applied all the permission on external location")

    def _update_location_permissions(self, location_name: str, principals: list[str]):
        permissions = [Privilege.CREATE_EXTERNAL_TABLE, Privilege.CREATE_EXTERNAL_VOLUME, Privilege.READ_FILES]
        changes = [PermissionsChange(add=permissions, principal=principal) for principal in principals]
        self._ws.grants.update(
            SecurableType.EXTERNAL_LOCATION,
            location_name,
            changes=changes,
        )

    def _get_location_name(self, location_url: str):
        for location in self._ws.external_locations.list():
            if location.url == location_url:
                return location.name
        return None



from dataclasses import dataclass
from typing import Any

from databricks.sdk.service._internal import _from_dict
from databricks.sdk.service.sql import Visualization, Widget, WidgetPosition

# this file is going away soon
# pylint: disable=redefined-builtin


@dataclass
class WidgetOptions:
    created_at: str | None = None
    description: str | None = None
    is_hidden: bool | None = None
    parameter_mappings: Any | None = None
    position: WidgetPosition | None = None
    title: str | None = None
    updated_at: str | None = None

    def as_dict(self) -> dict:
        body: dict[str, Any] = {}
        if self.created_at is not None:
            body["created_at"] = self.created_at
        if self.description is not None:
            body["description"] = self.description
        if self.is_hidden is not None:
            body["isHidden"] = self.is_hidden
        if self.parameter_mappings:
            body["parameterMappings"] = self.parameter_mappings
        if self.position:
            body["position"] = self.position.as_dict()
        if self.title is not None:
            body["title"] = self.title
        if self.updated_at is not None:
            body["updated_at"] = self.updated_at
        return body

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "WidgetOptions":
        return cls(
            created_at=data.get("created_at", None),
            description=data.get("description", None),
            is_hidden=data.get("isHidden", None),
            parameter_mappings=data.get("parameterMappings", None),
            position=_from_dict(data, "position", WidgetPosition),
            title=data.get("title", None),
            updated_at=data.get("updated_at", None),
        )


class DashboardWidgetsAPI:
    """This is an evolving API that facilitates the addition and removal of widgets from existing dashboards
    within the Databricks Workspace. Data structures may change over time."""

    def __init__(self, api_client):
        self._api = api_client

    def create(
        self,
        dashboard_id: str,
        options: WidgetOptions,
        *,
        text: str | None = None,
        visualization_id: str | None = None,
        width: int | None = None,
    ) -> Widget:
        """Add widget to a dashboard.

        :param dashboard_id: str
          Dashboard ID returned by :method:dashboards/create.
        :param options: :class:`WidgetOptions` (optional)
        :param text: str (optional)
          If this is a textbox widget, the application displays this text. This field is ignored if the widget
          contains a visualization in the `visualization` field.
        :param visualization_id: str (optional)
          Query Vizualization ID returned by :method:queryvisualizations/create.
        :param width: int (optional)
          Width of a widget

        :returns: :class:`Widget`
        """
        body: dict[str, str | dict | int] = {}
        if dashboard_id is not None:
            body["dashboard_id"] = dashboard_id
        if options is not None:
            body["options"] = options.as_dict()
        if text is not None:
            body["text"] = text
        if visualization_id is not None:
            body["visualization_id"] = visualization_id
        if width is not None:
            body["width"] = width
        res = self._api.do("POST", "/api/2.0/preview/sql/widgets", body=body)
        return Widget.from_dict(res)

    def delete(self, widget_id: str):
        self._api.do("DELETE", f"/api/2.0/preview/sql/widgets/{widget_id}")

    def update(
        self,
        dashboard_id: str,
        widget_id: str,
        *,
        options: WidgetOptions | None = None,
        text: str | None = None,
        visualization_id: str | None = None,
        width: int | None = None,
    ) -> Widget:
        """Update existing widget.

        :param dashboard_id: str
          Dashboard ID returned by :method:dashboards/create.
        :param widget_id: str
        :param options: :class:`WidgetOptions` (optional)
        :param text: str (optional)
          If this is a textbox widget, the application displays this text. This field is ignored if the widget
          contains a visualization in the `visualization` field.
        :param visualization_id: str (optional)
          Query Vizualization ID returned by :method:queryvisualizations/create.
        :param width: int (optional)
          Width of a widget

        :returns: :class:`Widget`
        """
        body: dict[str, str | dict | int] = {}
        if dashboard_id is not None:
            body["dashboard_id"] = dashboard_id
        if options is not None:
            body["options"] = options.as_dict()
        if text is not None:
            body["text"] = text
        if visualization_id is not None:
            body["visualization_id"] = visualization_id
        if width is not None:
            body["width"] = width
        res = self._api.do("POST", f"/api/2.0/preview/sql/widgets/{widget_id}", body=body)
        return Widget.from_dict(res)


class QueryVisualizationsAPI:
    """This is an evolving API that facilitates the addition and removal of vizualisations from existing queries
    within the Databricks Workspace. Data structures may change over time."""

    def __init__(self, api_client):
        self._api = api_client

    def create(
        self,
        query_id: str,
        type: str,
        options: dict,
        *,
        created_at: str | None = None,
        description: str | None = None,
        name: str | None = None,
        updated_at: str | None = None,
    ) -> Visualization:
        body: dict[str, Any] = {}
        if query_id is not None:
            body["query_id"] = query_id
        if type is not None:
            body["type"] = type
        if options is not None:
            body["options"] = options
        if name is not None:
            body["name"] = name
        if created_at is not None:
            body["created_at"] = created_at
        if description is not None:
            body["description"] = description
        if updated_at is not None:
            body["updated_at"] = updated_at
        res = self._api.do("POST", "/api/2.0/preview/sql/visualizations", body=body)
        return Visualization.from_dict(res)

    def delete(self, widget_id: str):
        """Remove visualization.

        :param widget_id: str
        """

        headers = {
            "Accept": "application/json",
        }
        self._api.do("DELETE", f"/api/2.0/preview/sql/visualizations/{widget_id}", headers=headers)



import json
import logging
from collections.abc import Callable, Iterable, Iterator, Sequence
from itertools import groupby

from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.crawlers import (
    CrawlerBase,
    Dataclass,
    DataclassInstance,
)
from databricks.labs.ucx.workspace_access.base import AclSupport, Permissions
from databricks.labs.ucx.workspace_access.groups import MigrationState

logger = logging.getLogger(__name__)


class PermissionManager(CrawlerBase[Permissions]):
    ERRORS_TO_IGNORE = ["FEATURE_DISABLED"]

    def __init__(self, backend: SqlBackend, inventory_database: str, crawlers: list[AclSupport]):
        super().__init__(backend, "hive_metastore", inventory_database, "permissions", Permissions)
        self._acl_support = crawlers

    def inventorize_permissions(self):
        # TODO: rename into snapshot()
        logger.debug("Crawling permissions")
        crawler_tasks = list(self._get_crawler_tasks())
        logger.info(f"Starting to crawl permissions. Total tasks: {len(crawler_tasks)}")
        items, errors = Threads.gather("crawl permissions", crawler_tasks)
        acute_errors = []
        for error in errors:
            if hasattr(error, 'error_code') and error.error_code in self.ERRORS_TO_IGNORE:
                logger.info(f"Error while crawling permissions: {error}. Skipping")
                continue
            logger.error(f"Error while crawling permissions: {error}")
            acute_errors.append(error)
        if len(acute_errors) > 0:
            raise ManyError(acute_errors)
        logger.info(f"Total crawled permissions: {len(items)}")
        self._save(items)
        logger.info(f"Saved {len(items)} to {self.full_name}")

    def apply_group_permissions(self, migration_state: MigrationState) -> bool:
        # list shall be sorted prior to using group by
        if len(migration_state) == 0:
            logger.info("No valid groups selected, nothing to do.")
            return True
        items = sorted(self.load_all(), key=lambda i: i.object_type)
        logger.info(
            f"Applying the permissions to account groups. "
            f"Total groups to apply permissions: {len(migration_state)}. "
            f"Total permissions found: {len(items)}"
        )
        applier_tasks: list[Callable[..., None]] = []
        supports_to_items = {
            support: list(items_subset) for support, items_subset in groupby(items, key=lambda i: i.object_type)
        }

        appliers = self.object_type_support()

        # we first check that all supports are valid.
        for object_type in supports_to_items:
            if object_type not in appliers:
                msg = f"Could not find support for {object_type}. Please check the inventory table."
                raise ValueError(msg)

        for object_type, items_subset in supports_to_items.items():
            relevant_support = appliers[object_type]
            tasks_for_support: list[Callable[..., None]] = []
            for item in items_subset:
                if not item:
                    continue
                task = relevant_support.get_apply_task(item, migration_state)
                if not task:
                    continue
                tasks_for_support.append(task)
            if len(tasks_for_support) == 0:
                continue
            logger.info(f"Total tasks for {object_type}: {len(tasks_for_support)}")
            applier_tasks.extend(tasks_for_support)

        logger.info(f"Starting to apply permissions on account groups. Total tasks: {len(applier_tasks)}")

        _, errors = Threads.gather("apply account group permissions", applier_tasks)
        if len(errors) > 0:
            logger.error(f"Detected {len(errors)} while applying permissions")
            raise ManyError(errors)
        logger.info("Permissions were applied")
        return True

    def verify_group_permissions(self) -> bool:
        items = sorted(self.load_all(), key=lambda i: i.object_type)
        logger.info(f"Total permissions found: {len(items)}")
        verifier_tasks: list[Callable[..., bool]] = []
        appliers = self.object_type_support()

        for object_type, items_subset in groupby(items, key=lambda i: i.object_type):
            if object_type not in appliers:
                msg = f"Could not find support for {object_type}. Please check the inventory table."
                raise ValueError(msg)

            relevant_support = appliers[object_type]
            tasks_for_support: list[Callable[..., bool]] = []
            for item in items_subset:
                task = relevant_support.get_verify_task(item)
                if not task:
                    continue
                tasks_for_support.append(task)

            logger.info(f"Total tasks for {object_type}: {len(tasks_for_support)}")
            verifier_tasks.extend(tasks_for_support)

        logger.info(f"Starting to verify permissions. Total tasks: {len(verifier_tasks)}")
        Threads.strict("verify group permissions", verifier_tasks)
        logger.info("All permissions validated successfully. No issues found.")

        return True

    def object_type_support(self) -> dict[str, AclSupport]:
        appliers: dict[str, AclSupport] = {}
        for support in self._acl_support:
            for object_type in support.object_types():
                if object_type in appliers:
                    msg = f"{object_type} is already supported by {type(appliers[object_type]).__name__}"
                    raise KeyError(msg)
                appliers[object_type] = support
        return appliers

    def cleanup(self):
        logger.info(f"Cleaning up inventory table {self.full_name}")
        self._exec(f"DROP TABLE IF EXISTS {self.full_name}")
        logger.info("Inventory table cleanup complete")

    def _save(self, items: Sequence[Permissions]):
        # keep in mind, that object_type and object_id are not primary keys.
        self._append_records(items)  # TODO: update instead of append
        logger.info("Successfully saved the items to inventory table")

    def load_all(self) -> list[Permissions]:
        logger.info(f"Loading inventory table {self.full_name}")
        if list(self._fetch(f"SELECT COUNT(*) as cnt FROM {self.full_name}"))[0][0] == 0:  # noqa: RUF015
            msg = (
                f"table {self.full_name} is empty for fetching permission info. "
                f"Please ensure assessment job is run successfully and permissions populated"
            )
            raise RuntimeError(msg)
        return [
            Permissions(object_id, object_type, raw)
            for object_id, object_type, raw in self._fetch(f"SELECT object_id, object_type, raw FROM {self.full_name}")
        ]

    def load_all_for(self, object_type: str, object_id: str, klass: Dataclass) -> Iterable[DataclassInstance]:
        for perm in self.load_all():
            if object_type == perm.object_type and object_id.lower() == perm.object_id.lower():
                raw = json.loads(perm.raw)
                yield klass(**raw)

    def _get_crawler_tasks(self) -> Iterator[Callable[..., Permissions | None]]:
        for support in self._acl_support:
            yield from support.get_crawler_tasks()



import dataclasses
import logging
import os
import sys
from pathlib import Path

from databricks.sdk.config import with_user_agent_extra

from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.assessment.workflows import Assessment, Failing
from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Task, Workflow, parse_args
from databricks.labs.ucx.installer.logs import TaskLogger
from databricks.labs.ucx.hive_metastore.workflows import (
    ScanTablesInMounts,
    TableMigration,
    MigrateTablesInMounts,
    MigrateHiveSerdeTablesInPlace,
    MigrateExternalTablesCTAS,
)
from databricks.labs.ucx.recon.workflows import MigrationRecon
from databricks.labs.ucx.source_code.workflows import ExperimentalWorkflowLinter
from databricks.labs.ucx.workspace_access.workflows import (
    GroupMigration,
    PermissionsMigrationAPI,
    RemoveWorkspaceLocalGroups,
    ValidateGroupPermissions,
)

logger = logging.getLogger(__name__)


class Workflows:
    def __init__(self, workflows: list[Workflow]):
        self._tasks: list[Task] = []
        self._workflows: dict[str, Workflow] = {}
        for workflow in workflows:
            self._workflows[workflow.name] = workflow
            for task_definition in workflow.tasks():
                # Add the workflow name to the task definition, because we cannot access
                # the workflow name from the method decorator
                with_workflow = dataclasses.replace(task_definition, workflow=workflow.name)
                self._tasks.append(with_workflow)

    @classmethod
    def all(cls):
        return cls(
            [
                Assessment(),
                GroupMigration(),
                TableMigration(),
                MigrateHiveSerdeTablesInPlace(),
                MigrateExternalTablesCTAS(),
                ValidateGroupPermissions(),
                RemoveWorkspaceLocalGroups(),
                ScanTablesInMounts(),
                MigrateTablesInMounts(),
                PermissionsMigrationAPI(),
                ExperimentalWorkflowLinter(),
                MigrationRecon(),
                Failing(),
            ]
        )

    def tasks(self) -> list[Task]:
        return self._tasks

    def trigger(self, *argv):
        named_parameters = parse_args(*argv)
        config_path = Path(named_parameters["config"])
        ctx = RuntimeContext(named_parameters)
        install_dir = config_path.parent
        task_name = named_parameters.get("task", "not specified")
        workflow_name = named_parameters.get("workflow", "not specified")
        attempt = named_parameters.get("attempt", "0")
        if workflow_name not in self._workflows:
            msg = f'workflow "{workflow_name}" not found. Valid workflows are: {", ".join(self._workflows.keys())}'
            raise KeyError(msg)
        print(f"UCX v{__version__}")
        workflow = self._workflows[workflow_name]
        if task_name == "parse_logs":
            return ctx.task_run_warning_recorder.snapshot()
        # both CLI commands and workflow names appear in telemetry under `cmd`
        with_user_agent_extra("cmd", workflow_name)
        # `{{parent_run_id}}` is the run of entire workflow, whereas `{{run_id}}` is the run of a task
        workflow_run_id = named_parameters.get("parent_run_id", "unknown_run_id")
        job_id = named_parameters.get("job_id", "unknown_job_id")
        with TaskLogger(
            install_dir,
            workflow=workflow_name,
            workflow_id=job_id,
            task_name=task_name,
            workflow_run_id=workflow_run_id,
            log_level=ctx.config.log_level,
            attempt=attempt,
        ) as task_logger:
            ucx_logger = logging.getLogger("databricks.labs.ucx")
            ucx_logger.info(f"UCX v{__version__} After job finishes, see debug logs at {task_logger}")
            current_task = getattr(workflow, task_name)
            current_task(ctx)
            return None


def main(*argv):
    if len(argv) == 0:
        argv = sys.argv
    Workflows.all().trigger(*argv)


if __name__ == "__main__":
    if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
        raise SystemExit("Only intended to run in Databricks Runtime")
    main(*sys.argv)



from abc import abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass

from astroid import Attribute, Call, Name, NodeNG  # type: ignore
from databricks.labs.ucx.source_code.base import (
    Advice,
    Failure,
    PythonLinter,
)
from databricks.labs.ucx.source_code.linters.python_ast import Tree


@dataclass
class SharedClusterMatcher:
    is_serverless: bool

    def _cluster_type_str(self) -> str:
        return 'UC Shared Clusters' if not self.is_serverless else 'Serverless Compute'

    @abstractmethod
    def lint(self, node: NodeNG) -> Iterator[Advice]:
        pass

    def lint_tree(self, root: NodeNG) -> Iterator[Advice]:
        reported_locations = set()
        for node in Tree(root).walk():
            for advice in self.lint(node):
                loc = (advice.start_line, advice.start_col)
                if loc not in reported_locations:
                    reported_locations.add(loc)
                    yield advice


class JvmAccessMatcher(SharedClusterMatcher):
    _FIELDS = [
        "_jvm",
        "_jcol",
        "_jdf",
        "_jspark",
        "_jsparkSession",
    ]

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return
        if node.attrname not in JvmAccessMatcher._FIELDS:
            return
        yield Failure.from_node(
            code='jvm-access-in-shared-clusters',
            message=f'Cannot access Spark Driver JVM on {self._cluster_type_str()}',
            node=node,
        )


class RDDApiMatcher(SharedClusterMatcher):
    _SC_METHODS = [
        # RDD creation
        "emptyRDD",
        "parallelize",
        "range",
        "makeRDD",
        # file read methods
        "binaryFiles",
        "binaryRecords",
        "hadoopFile",
        "hadoopRDD",
        "newAPIHadoopFile",
        "newAPIHadoopRDD",
        "objectFile",
        "sequenceFile",
        "textFile",
        "wholeTextFiles",
    ]

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        yield from self._lint_sc(node)
        yield from self._lint_rdd_use(node)

    def _lint_rdd_use(self, node: NodeNG) -> Iterator[Advice]:
        if isinstance(node, Attribute):
            if node.attrname == 'rdd':
                yield self._rdd_failure(node)
                return
        if isinstance(node, Call) and isinstance(node.func, Attribute) and node.func.attrname == 'mapPartitions':
            yield Failure.from_node(
                code='rdd-in-shared-clusters',
                message=f'RDD APIs are not supported on {self._cluster_type_str()}. '
                f'Use mapInArrow() or Pandas UDFs instead',
                node=node,
            )

    def _lint_sc(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Call) or not isinstance(node.func, Attribute):
            return
        if node.func.attrname not in self._SC_METHODS:
            return
        function_name = Tree.get_full_function_name(node)
        if not function_name or not function_name.endswith(f"sc.{node.func.attrname}"):
            return
        yield self._rdd_failure(node)

    def _rdd_failure(self, node: NodeNG) -> Advice:
        return Failure.from_node(
            code='rdd-in-shared-clusters',
            message=f'RDD APIs are not supported on {self._cluster_type_str()}. Rewrite it using DataFrame API',
            node=node,
        )


class SparkSqlContextMatcher(SharedClusterMatcher):
    _ATTRIBUTES = ["sc", "sqlContext", "sparkContext"]
    _KNOWN_REPLACEMENTS = {"getConf": "conf", "_conf": "conf"}

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return

        if isinstance(node.expr, Name) and node.expr.name in SparkSqlContextMatcher._ATTRIBUTES:
            yield self._get_advice(node, node.expr.name)
        # sparkContext can be an attribute as in df.sparkContext.getConf()
        if isinstance(node.expr, Attribute) and node.expr.attrname == 'sparkContext':
            yield self._get_advice(node, node.expr.attrname)

    def _get_advice(self, node: Attribute, name: str) -> Advice:
        if node.attrname in self._KNOWN_REPLACEMENTS:
            replacement = self._KNOWN_REPLACEMENTS[node.attrname]
            return Failure.from_node(
                code='legacy-context-in-shared-clusters',
                message=f'{name} and {node.attrname} are not supported on {self._cluster_type_str()}. '
                f'Rewrite it using spark.{replacement}',
                node=node,
            )
        return Failure.from_node(
            code='legacy-context-in-shared-clusters',
            message=f'{name} is not supported on {self._cluster_type_str()}. Rewrite it using spark',
            node=node,
        )


class LoggingMatcher(SharedClusterMatcher):
    def lint(self, node: NodeNG) -> Iterator[Advice]:
        yield from self._match_sc_set_log_level(node)
        yield from self._match_jvm_log(node)

    def _match_sc_set_log_level(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Call) or not isinstance(node.func, Attribute):
            return
        if node.func.attrname != 'setLogLevel':
            return
        function_name = Tree.get_full_function_name(node)
        if not function_name or not function_name.endswith('sc.setLogLevel'):
            return

        yield Failure.from_node(
            code='spark-logging-in-shared-clusters',
            message=f'Cannot set Spark log level directly from code on {self._cluster_type_str()}. '
            f'Remove the call and set the cluster spark conf \'spark.log.level\' instead',
            node=node,
        )

    def _match_jvm_log(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return
        attribute_name = Tree.get_full_attribute_name(node)
        if attribute_name and attribute_name.endswith('org.apache.log4j'):
            yield Failure.from_node(
                code='spark-logging-in-shared-clusters',
                message=f'Cannot access Spark Driver JVM logger on {self._cluster_type_str()}. '
                f'Use logging.getLogger() instead',
                node=node,
            )


class SparkConnectLinter(PythonLinter):
    def __init__(self, is_serverless: bool = False):
        self._matchers = [
            JvmAccessMatcher(is_serverless=is_serverless),
            RDDApiMatcher(is_serverless=is_serverless),
            SparkSqlContextMatcher(is_serverless=is_serverless),
            LoggingMatcher(is_serverless=is_serverless),
        ]

    def lint_tree(self, tree: Tree) -> Iterator[Advice]:
        for matcher in self._matchers:
            yield from matcher.lint_tree(tree.node)



from functools import partial

from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.catalog import (
    PermissionsChange,
    Privilege,
    SecurableType,
    TableType,
)

from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.table_migrate import logger


class TableMove:
    def __init__(self, ws: WorkspaceClient, backend: SqlBackend):
        self._backend = backend
        self._fetch = backend.fetch
        self._execute = backend.execute
        self._ws = ws

    def move(
        self,
        from_catalog: str,
        from_schema: str,
        from_table: str,
        to_catalog: str,
        to_schema: str,
        *,
        del_table: bool,
    ):
        try:
            self._ws.schemas.get(f"{from_catalog}.{from_schema}")
        except NotFound:
            logger.error(f"schema {from_schema} not found in catalog {from_catalog}, enter correct schema details.")
            return
        try:
            self._ws.schemas.get(f"{to_catalog}.{to_schema}")
        except NotFound:
            logger.warning(f"schema {to_schema} not found in {to_catalog}, creating...")
            self._ws.schemas.create(to_schema, to_catalog)

        tables = self._ws.tables.list(from_catalog, from_schema)
        table_tasks = []
        view_tasks = []
        filtered_tables = [table for table in tables if from_table in [table.name, "*"]]
        for table in filtered_tables:
            try:
                self._ws.tables.get(f"{to_catalog}.{to_schema}.{table.name}")
                logger.warning(
                    f"table {from_table} already present in {to_catalog}.{to_schema}. skipping this table..."
                )
                continue
            except NotFound:
                if table.table_type and table.table_type in (TableType.EXTERNAL, TableType.MANAGED):
                    table_tasks.append(
                        partial(
                            self._move_table,
                            from_catalog,
                            from_schema,
                            table.name,
                            to_catalog,
                            to_schema,
                            table.table_type,
                            del_table=del_table,
                        )
                    )
                    continue
                if table.view_definition:
                    view_tasks.append(
                        partial(
                            self._move_view,
                            from_catalog,
                            from_schema,
                            table.name,
                            to_catalog,
                            to_schema,
                            del_view=del_table,
                            view_text=table.view_definition,
                        )
                    )
                    continue
                logger.warning(
                    f"table {from_table} was not identified as a valid table or view. skipping this table..."
                )
        Threads.strict("Creating tables", table_tasks)
        logger.info(f"Moved {len(list(table_tasks))} tables to the new schema {to_schema}.")
        Threads.strict("Creating views", view_tasks)
        logger.info(f"Moved {len(list(view_tasks))} views to the new schema {to_schema}.")

    def alias_tables(
        self,
        from_catalog: str,
        from_schema: str,
        from_table: str,
        to_catalog: str,
        to_schema: str,
    ):
        try:
            self._ws.schemas.get(f"{from_catalog}.{from_schema}")
        except NotFound:
            logger.error(f"schema {from_schema} not found in catalog {from_catalog}, enter correct schema details.")
            return
        try:
            self._ws.schemas.get(f"{to_catalog}.{to_schema}")
        except NotFound:
            logger.warning(f"schema {to_schema} not found in {to_catalog}, creating...")
            self._ws.schemas.create(to_schema, to_catalog)

        tables = self._ws.tables.list(from_catalog, from_schema)
        alias_tasks = []
        filtered_tables = [table for table in tables if from_table in [table.name, "*"]]
        for table in filtered_tables:
            try:
                self._ws.tables.get(f"{to_catalog}.{to_schema}.{table.name}")
                logger.warning(
                    f"table {from_table} already present in {from_catalog}.{from_schema}. skipping this table..."
                )
                continue
            except NotFound:
                if table.table_type and table.table_type in (TableType.EXTERNAL, TableType.MANAGED):
                    alias_tasks.append(
                        partial(self._alias_table, from_catalog, from_schema, table.name, to_catalog, to_schema)
                    )
                    continue
                if table.view_definition:
                    alias_tasks.append(
                        partial(
                            self._move_view,
                            from_catalog,
                            from_schema,
                            table.name,
                            to_catalog,
                            to_schema,
                            del_view=False,
                            view_text=table.view_definition,
                        )
                    )
                    continue
                logger.warning(
                    f"table {from_table} was not identified as a valid table or view. skipping this table..."
                )
        Threads.strict("Creating aliases", alias_tasks)
        logger.info(f"Created {len(list(alias_tasks))} table and view aliases in the new schema {to_schema}.")

    def _move_table(
        self,
        from_catalog: str,
        from_schema: str,
        from_table: str,
        to_catalog: str,
        to_schema: str,
        table_type: str,
        *,
        del_table: bool,
    ) -> bool:
        from_table_name = f"{from_catalog}.{from_schema}.{from_table}"
        to_table_name = f"{to_catalog}.{to_schema}.{from_table}"
        try:
            self._recreate_table(
                from_table_name, to_table_name, is_managed=table_type == TableType.MANAGED, del_table=del_table
            )
            self._reapply_grants(from_table_name, to_table_name)
        except NotFound as err:
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(err) or "[DELTA_TABLE_NOT_FOUND]" in str(err):
                logger.error(f"Could not find table {from_table_name}. Table not found.")
            else:
                logger.error(f"Failed to move table {from_table_name}: {err!s}", exc_info=True)
        return False

    def _alias_table(
        self,
        from_catalog: str,
        from_schema: str,
        from_table: str,
        to_catalog: str,
        to_schema: str,
    ) -> bool:
        from_table_name = f"{from_catalog}.{from_schema}.{from_table}"
        to_table_name = f"{to_catalog}.{to_schema}.{from_table}"
        try:
            self._create_alias_view(from_table_name, to_table_name)
            self._reapply_grants(from_table_name, to_table_name, target_view=True)
            return True
        except NotFound as err:
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(err) or "[DELTA_TABLE_NOT_FOUND]" in str(err):
                logger.error(f"Could not find table {from_table_name}. Table not found.")
            else:
                logger.error(f"Failed to alias table {from_table_name}: {err!s}", exc_info=True)
            return False

    def _reapply_grants(self, from_table_name, to_table_name, *, target_view: bool = False):
        try:
            grants = self._ws.grants.get(SecurableType.TABLE, from_table_name)
        except NotFound:
            logger.warning(f"removed on the backend {from_table_name}")
            return
        if not grants.privilege_assignments:
            return
        logger.info(f"Applying grants on table {to_table_name}")
        grants_changes = []
        for permission in grants.privilege_assignments:
            if not permission.privileges:
                continue
            if not target_view:
                grants_changes.append(PermissionsChange(list(permission.privileges), permission.principal))
                continue
            privileges = set()
            for privilege in permission.privileges:
                if privilege != Privilege.MODIFY:
                    privileges.add(privilege)
            if privileges:
                grants_changes.append(PermissionsChange(list(privileges), permission.principal))

        self._ws.grants.update(SecurableType.TABLE, to_table_name, changes=grants_changes)

    def _recreate_table(self, from_table_name, to_table_name, *, is_managed: bool, del_table: bool):
        drop_table = f"DROP TABLE {escape_sql_identifier(from_table_name)}"
        if is_managed:
            create_table_sql = (
                f"CREATE TABLE IF NOT EXISTS {escape_sql_identifier(to_table_name)} "
                f"DEEP CLONE {escape_sql_identifier(from_table_name)}"
            )
            logger.info(f"Creating table {to_table_name}")
            self._execute(create_table_sql)
            if del_table:
                row_check = str(
                    next(
                        self._fetch(
                            f"SELECT (SELECT COUNT(*) FROM {escape_sql_identifier(to_table_name)})="
                            f"(SELECT COUNT(*) FROM {escape_sql_identifier(from_table_name)})"
                        )
                    )[0]
                )
                if row_check.lower() == "false":
                    logger.error(f"Number of rows did not match after cloning {from_table_name} to {to_table_name}.")
                    return
                logger.info(f"Dropping source table {from_table_name}")
                self._execute(drop_table)
        else:
            create_sql = str(next(self._fetch(f"SHOW CREATE TABLE {escape_sql_identifier(from_table_name)}"))[0])
            logger.info(f"Dropping source table {from_table_name}")
            self._execute(drop_table)
            create_table_sql = create_sql.replace(
                f"CREATE TABLE {escape_sql_identifier(from_table_name)}",
                f"CREATE TABLE {escape_sql_identifier(to_table_name)}",
            )
            logger.info(f"Creating table {to_table_name}")
            self._execute(create_table_sql)

    def _create_alias_view(self, from_table_name, to_table_name):
        create_view_sql = (
            f"CREATE VIEW {escape_sql_identifier(to_table_name)} "
            f"AS SELECT * FROM {escape_sql_identifier(from_table_name)}"
        )
        logger.info(f"Creating view {to_table_name} on {from_table_name}")
        self._execute(create_view_sql)

    def _move_view(
        self,
        from_catalog: str,
        from_schema: str,
        from_view: str,
        to_catalog: str,
        to_schema: str,
        *,
        del_view: bool,
        view_text: str | None = None,
    ) -> bool:
        from_view_name = f"{from_catalog}.{from_schema}.{from_view}"
        to_view_name = f"{to_catalog}.{to_schema}.{from_view}"
        try:
            self._recreate_view(to_view_name, view_text)
            self._reapply_grants(from_view_name, to_view_name)
            if del_view:
                logger.info(f"Dropping source view {from_view_name}")
                drop_sql = f"DROP VIEW {escape_sql_identifier(from_view_name)}"
                self._execute(drop_sql)
            return True
        except NotFound as err:
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(err):
                logger.error(f"Could not find view {from_view_name}. View not found.")
            else:
                logger.error(f"Failed to move view {from_view_name}: {err!s}", exc_info=True)
        return False

    def _recreate_view(self, to_view_name, view_text):
        create_sql = f"CREATE VIEW {escape_sql_identifier(to_view_name)} AS {view_text}"
        logger.info(f"Creating view {to_view_name}")
        self._execute(create_sql)



import logging
import os

from databricks.labs.blueprint.installation import Installation
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.sql import EndpointInfoWarehouseType

from databricks.labs.ucx.config import WorkspaceConfig

logger = logging.getLogger(__name__)


class InstallationMixin:
    def __init__(self, config: WorkspaceConfig, installation: Installation, ws: WorkspaceClient):
        self._config = config
        self._installation = installation
        self._ws = ws

    def _name(self, name: str) -> str:
        prefix = os.path.basename(self._installation.install_folder()).removeprefix('.')
        return f"[{prefix.upper()}] {name}"

    @property
    def _my_username(self):
        if not hasattr(self, "_me"):
            self._me = self._ws.current_user.me()
            is_workspace_admin = any(g.display == "admins" for g in self._me.groups)
            if not is_workspace_admin:
                msg = "Current user is not a workspace admin"
                raise PermissionError(msg)
        return self._me.user_name

    @property
    def _warehouse_id(self) -> str:
        if self._config.warehouse_id is not None:
            logger.info("Fetching warehouse_id from a config")
            return self._config.warehouse_id
        warehouses = [_ for _ in self._ws.warehouses.list() if _.warehouse_type == EndpointInfoWarehouseType.PRO]
        warehouse_id = self._config.warehouse_id
        if not warehouse_id and not warehouses:
            msg = "need either configured warehouse_id or an existing PRO SQL warehouse"
            raise ValueError(msg)
        if not warehouse_id:
            warehouse_id = warehouses[0].id
        self._config.warehouse_id = warehouse_id
        return warehouse_id



import io
import json
import logging
import os
import pathlib
import shutil
import string
import subprocess
import sys
from collections.abc import Callable, Generator, MutableMapping
from datetime import timedelta, datetime
from pathlib import Path
from typing import BinaryIO

import pytest
from databricks.labs.lsql.backends import StatementExecutionBackend
from databricks.labs.blueprint.commands import CommandExecutor
from databricks.sdk import AccountClient, WorkspaceClient
from databricks.sdk.core import DatabricksError
from databricks.sdk.errors import NotFound, ResourceConflict
from databricks.sdk.retries import retried
from databricks.sdk.service import compute, iam, jobs, pipelines, sql, workspace
from databricks.sdk.service._internal import Wait
from databricks.sdk.service.catalog import (
    AwsIamRoleRequest,
    AzureServicePrincipal,
    CatalogInfo,
    DataSourceFormat,
    FunctionInfo,
    SchemaInfo,
    StorageCredentialInfo,
    TableInfo,
    TableType,
)
from databricks.sdk.service.serving import (
    EndpointCoreConfigInput,
    ServedModelInput,
    ServedModelInputWorkloadSize,
    ServingEndpointDetailed,
)
from databricks.sdk.service.sql import (
    CreateWarehouseRequestWarehouseType,
    GetResponse,
    ObjectTypePlural,
    Query,
    Dashboard,
    WidgetOptions,
    WidgetPosition,
    EndpointTagPair,
    EndpointTags,
)
from databricks.sdk.service.workspace import ImportFormat, Language

from databricks.labs.ucx.workspace_access.groups import MigratedGroup

# this file will get to databricks-labs-pytester project and be maintained/refactored there
# pylint: disable=redefined-outer-name,too-many-try-statements,import-outside-toplevel,unnecessary-lambda,too-complex,invalid-name

logger = logging.getLogger(__name__)
TEST_JOBS_PURGE_TIMEOUT = timedelta(hours=1, minutes=15)  # 15 minutes grace for jobs starting at the end of the hour


def factory(name, create, remove):
    cleanup = []

    def inner(**kwargs):
        x = create(**kwargs)
        logger.debug(f"added {name} fixture: {x}")
        cleanup.append(x)
        return x

    yield inner
    logger.debug(f"clearing {len(cleanup)} {name} fixtures")
    for x in cleanup:
        try:
            logger.debug(f"removing {name} fixture: {x}")
            remove(x)
        except DatabricksError as e:
            # TODO: fix on the databricks-labs-pytester level
            logger.debug(f"ignoring error while {name} {x} teardown: {e}")


@pytest.fixture
def fresh_wheel_file(tmp_path) -> Path:
    this_file = Path(__file__)
    project_root = this_file.parent.parent.parent.parent.parent.parent.absolute()
    # TODO: we can dynamically determine this with python -m build .
    wheel_name = "databricks_labs_ucx"

    build_root = tmp_path / fresh_wheel_file.__name__
    shutil.copytree(project_root, build_root)
    try:
        completed_process = subprocess.run(
            [sys.executable, "-m", "pip", "wheel", "."],
            capture_output=True,
            cwd=build_root,
            check=True,
        )
        if completed_process.returncode != 0:
            raise RuntimeError(completed_process.stderr)

        found_wheels = list(build_root.glob(f"{wheel_name}-*.whl"))
        if not found_wheels:
            msg = f"cannot find {wheel_name}-*.whl"
            raise RuntimeError(msg)
        if len(found_wheels) > 1:
            conflicts = ", ".join(str(whl) for whl in found_wheels)
            msg = f"more than one wheel match: {conflicts}"
            raise RuntimeError(msg)
        wheel_file = found_wheels[0]

        return wheel_file
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr) from None


@pytest.fixture
def wsfs_wheel(ws, fresh_wheel_file, make_random):
    my_user = ws.current_user.me().user_name
    workspace_location = f"/Users/{my_user}/wheels/{make_random(10)}"
    ws.workspace.mkdirs(workspace_location)

    wsfs_wheel = f"{workspace_location}/{fresh_wheel_file.name}"
    with fresh_wheel_file.open("rb") as f:
        ws.workspace.upload(wsfs_wheel, f, format=ImportFormat.AUTO)

    yield wsfs_wheel

    ws.workspace.delete(workspace_location, recursive=True)


@pytest.fixture
def make_random() -> Callable[[int], str]:
    import random

    def inner(k=16) -> str:
        charset = string.ascii_uppercase + string.ascii_lowercase + string.digits
        return "".join(random.choices(charset, k=int(k)))

    return inner


@pytest.fixture
def product_info():
    return None, None


@pytest.fixture
def ws(product_info, debug_env) -> WorkspaceClient:
    # Use variables from Unified Auth
    # See https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html
    product_name, product_version = product_info
    return WorkspaceClient(host=debug_env["DATABRICKS_HOST"], product=product_name, product_version=product_version)


@pytest.fixture
def acc(product_info, debug_env) -> AccountClient:
    # Use variables from Unified Auth
    # See https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html
    product_name, product_version = product_info
    logger.debug(f"Running with {len(debug_env)} env variables")
    return AccountClient(
        host=debug_env["DATABRICKS_HOST"],
        account_id=debug_env["DATABRICKS_ACCOUNT_ID"],
        product=product_name,
        product_version=product_version,
    )


def _permissions_mapping():
    from databricks.sdk.service.iam import PermissionLevel

    def _simple(_, object_id):
        return object_id

    def _path(ws, path):
        return ws.workspace.get_status(path).object_id

    return [
        ("cluster_policy", "cluster-policies", [PermissionLevel.CAN_USE], _simple),
        (
            "instance_pool",
            "instance-pools",
            [PermissionLevel.CAN_ATTACH_TO, PermissionLevel.CAN_MANAGE],
            _simple,
        ),
        (
            "cluster",
            "clusters",
            [PermissionLevel.CAN_ATTACH_TO, PermissionLevel.CAN_RESTART, PermissionLevel.CAN_MANAGE],
            _simple,
        ),
        (
            "pipeline",
            "pipelines",
            [
                PermissionLevel.CAN_VIEW,
                PermissionLevel.CAN_RUN,
                PermissionLevel.CAN_MANAGE,
                PermissionLevel.IS_OWNER,  # cannot be a group
            ],
            _simple,
        ),
        (
            "job",
            "jobs",
            [
                PermissionLevel.CAN_VIEW,
                PermissionLevel.CAN_MANAGE_RUN,
                PermissionLevel.CAN_MANAGE,
                PermissionLevel.IS_OWNER,  # cannot be a group
            ],
            _simple,
        ),
        (
            "notebook",
            "notebooks",
            [PermissionLevel.CAN_READ, PermissionLevel.CAN_RUN, PermissionLevel.CAN_EDIT, PermissionLevel.CAN_MANAGE],
            _path,
        ),
        (
            "directory",
            "directories",
            [PermissionLevel.CAN_READ, PermissionLevel.CAN_RUN, PermissionLevel.CAN_EDIT, PermissionLevel.CAN_MANAGE],
            _path,
        ),
        (
            "workspace_file",
            "files",
            [PermissionLevel.CAN_READ, PermissionLevel.CAN_RUN, PermissionLevel.CAN_EDIT, PermissionLevel.CAN_MANAGE],
            _simple,
        ),
        (
            "workspace_file_path",
            "files",
            [PermissionLevel.CAN_READ, PermissionLevel.CAN_RUN, PermissionLevel.CAN_EDIT, PermissionLevel.CAN_MANAGE],
            _path,
        ),
        (
            "repo",
            "repos",
            [PermissionLevel.CAN_READ, PermissionLevel.CAN_RUN, PermissionLevel.CAN_EDIT, PermissionLevel.CAN_MANAGE],
            _path,
        ),
        ("authorization", "authorization", [PermissionLevel.CAN_USE], _simple),
        (
            "warehouse",
            "sql/warehouses",
            [PermissionLevel.CAN_USE, PermissionLevel.CAN_MANAGE],
            _simple,
        ),
        (
            "dashboard",
            "sql/dashboards",
            [PermissionLevel.CAN_EDIT, PermissionLevel.CAN_RUN, PermissionLevel.CAN_MANAGE, PermissionLevel.CAN_VIEW],
            _simple,
        ),
        (
            "alert",
            "sql/alerts",
            [PermissionLevel.CAN_EDIT, PermissionLevel.CAN_RUN, PermissionLevel.CAN_MANAGE, PermissionLevel.CAN_VIEW],
            _simple,
        ),
        (
            "query",
            "sql/queries",
            [PermissionLevel.CAN_EDIT, PermissionLevel.CAN_RUN, PermissionLevel.CAN_MANAGE, PermissionLevel.CAN_VIEW],
            _simple,
        ),
        (
            "experiment",
            "experiments",
            [PermissionLevel.CAN_READ, PermissionLevel.CAN_EDIT, PermissionLevel.CAN_MANAGE],
            _simple,
        ),
        (
            "registered_model",
            "registered-models",
            [
                PermissionLevel.CAN_READ,
                PermissionLevel.CAN_EDIT,
                PermissionLevel.CAN_MANAGE_STAGING_VERSIONS,
                PermissionLevel.CAN_MANAGE_PRODUCTION_VERSIONS,
                PermissionLevel.CAN_MANAGE,
            ],
            _simple,
        ),
        (
            "serving_endpoint",
            "serving-endpoints",
            [PermissionLevel.CAN_VIEW, PermissionLevel.CAN_MANAGE, PermissionLevel.CAN_QUERY],
            _simple,
        ),
        (
            "feature_table",
            "feature-tables",
            [PermissionLevel.CAN_VIEW_METADATA, PermissionLevel.CAN_EDIT_METADATA, PermissionLevel.CAN_MANAGE],
            _simple,
        ),
    ]


def _redash_permissions_mapping():
    def _simple(_, object_id):
        return object_id

    return [
        (
            "query",
            ObjectTypePlural.QUERIES,
            [
                sql.PermissionLevel.CAN_VIEW,
                sql.PermissionLevel.CAN_RUN,
                sql.PermissionLevel.CAN_MANAGE,
                sql.PermissionLevel.CAN_EDIT,
            ],
            _simple,
        ),
        (
            "alert",
            ObjectTypePlural.ALERTS,
            [
                sql.PermissionLevel.CAN_VIEW,
                sql.PermissionLevel.CAN_RUN,
                sql.PermissionLevel.CAN_MANAGE,
                sql.PermissionLevel.CAN_EDIT,
            ],
            _simple,
        ),
        (
            "dashboard",
            ObjectTypePlural.DASHBOARDS,
            [
                sql.PermissionLevel.CAN_VIEW,
                sql.PermissionLevel.CAN_RUN,
                sql.PermissionLevel.CAN_MANAGE,
                sql.PermissionLevel.CAN_EDIT,
            ],
            _simple,
        ),
    ]


class _PermissionsChange:
    def __init__(self, object_id: str, before: list[iam.AccessControlRequest], after: list[iam.AccessControlRequest]):
        self.object_id = object_id
        self.before = before
        self.after = after

    @staticmethod
    def _principal(acr: iam.AccessControlRequest) -> str:
        if acr.user_name is not None:
            return f"user_name {acr.user_name}"
        if acr.group_name is not None:
            return f"group_name {acr.group_name}"
        return f"service_principal_name {acr.service_principal_name}"

    def _list(self, acl: list[iam.AccessControlRequest]):
        return ", ".join(f"{self._principal(_)} {_.permission_level.value}" for _ in acl)

    def __repr__(self):
        return f"{self.object_id} [{self._list(self.before)}] -> [{self._list(self.after)}]"


class _RedashPermissionsChange:
    def __init__(self, object_id: str, before: list[sql.AccessControl], after: list[sql.AccessControl]):
        self.object_id = object_id
        self.before = before
        self.after = after

    @staticmethod
    def _principal(acr: sql.AccessControl) -> str:
        if acr.user_name is not None:
            return f"user_name {acr.user_name}"
        return f"group_name {acr.group_name}"

    def _list(self, acl: list[sql.AccessControl]):
        return ", ".join(f"{self._principal(_)} {_.permission_level.value}" for _ in acl)

    def __repr__(self):
        return f"{self.object_id} [{self._list(self.before)}] -> [{self._list(self.after)}]"


def _make_permissions_factory(name, resource_type, levels, id_retriever):
    def _non_inherited(x: iam.ObjectPermissions):
        out: list[iam.AccessControlRequest] = []
        assert x.access_control_list is not None
        for access_control in x.access_control_list:
            if not access_control.all_permissions:
                continue
            for permission in access_control.all_permissions:
                if not permission.inherited:
                    continue
                out.append(
                    iam.AccessControlRequest(
                        permission_level=permission.permission_level,
                        group_name=access_control.group_name,
                        user_name=access_control.user_name,
                        service_principal_name=access_control.service_principal_name,
                    )
                )
        return out

    def _make_permissions(ws):
        def create(
            *,
            object_id: str,
            permission_level: iam.PermissionLevel | None = None,
            group_name: str | None = None,
            user_name: str | None = None,
            service_principal_name: str | None = None,
            access_control_list: list[iam.AccessControlRequest] | None = None,
        ):
            nothing_specified = permission_level is None and access_control_list is None
            both_specified = permission_level is not None and access_control_list is not None
            if nothing_specified or both_specified:
                msg = "either permission_level or access_control_list has to be specified"
                raise ValueError(msg)

            object_id = id_retriever(ws, object_id)
            initial = _non_inherited(ws.permissions.get(resource_type, object_id))
            if access_control_list is None:
                if permission_level not in levels:
                    assert permission_level is not None
                    names = ", ".join(_.value for _ in levels)
                    msg = f"invalid permission level: {permission_level.value}. Valid levels: {names}"
                    raise ValueError(msg)

                access_control_list = []
                if group_name is not None:
                    access_control_list.append(
                        iam.AccessControlRequest(
                            group_name=group_name,
                            permission_level=permission_level,
                        )
                    )
                if user_name is not None:
                    access_control_list.append(
                        iam.AccessControlRequest(
                            user_name=user_name,
                            permission_level=permission_level,
                        )
                    )
                if service_principal_name is not None:
                    access_control_list.append(
                        iam.AccessControlRequest(
                            service_principal_name=service_principal_name,
                            permission_level=permission_level,
                        )
                    )
            ws.permissions.update(resource_type, object_id, access_control_list=access_control_list)
            return _PermissionsChange(object_id, initial, access_control_list)

        def remove(change: _PermissionsChange):
            ws.permissions.set(resource_type, change.object_id, access_control_list=change.before)

        yield from factory(f"{name} permissions", create, remove)

    return _make_permissions


def _make_redash_permissions_factory(name, resource_type, levels, id_retriever):
    def _non_inherited(x: GetResponse):
        out: list[sql.AccessControl] = []
        assert x.access_control_list is not None
        for access_control in x.access_control_list:
            out.append(
                sql.AccessControl(
                    permission_level=access_control.permission_level,
                    group_name=access_control.group_name,
                    user_name=access_control.user_name,
                )
            )
        return out

    def _make_permissions(ws):
        def create(
            *,
            object_id: str,
            permission_level: sql.PermissionLevel | None = None,
            group_name: str | None = None,
            user_name: str | None = None,
            access_control_list: list[sql.AccessControl] | None = None,
        ):
            nothing_specified = permission_level is None and access_control_list is None
            both_specified = permission_level is not None and access_control_list is not None
            if nothing_specified or both_specified:
                msg = "either permission_level or access_control_list has to be specified"
                raise ValueError(msg)

            object_id = id_retriever(ws, object_id)
            initial = _non_inherited(ws.dbsql_permissions.get(resource_type, object_id))

            if access_control_list is None:
                if permission_level not in levels:
                    assert permission_level is not None
                    names = ", ".join(_.value for _ in levels)
                    msg = f"invalid permission level: {permission_level.value}. Valid levels: {names}"
                    raise ValueError(msg)

                access_control_list = []
                if group_name is not None:
                    access_control_list.append(
                        sql.AccessControl(
                            group_name=group_name,
                            permission_level=permission_level,
                        )
                    )
                if user_name is not None:
                    access_control_list.append(
                        sql.AccessControl(
                            user_name=user_name,
                            permission_level=permission_level,
                        )
                    )

            ws.dbsql_permissions.set(resource_type, object_id, access_control_list=access_control_list)
            return _RedashPermissionsChange(object_id, initial, access_control_list)

        def remove(change: _RedashPermissionsChange):
            ws.dbsql_permissions.set(
                sql.ObjectTypePlural(resource_type), change.object_id, access_control_list=change.before
            )

        yield from factory(f"{name} permissions", create, remove)

    return _make_permissions


for name, resource_type, levels, id_retriever in _permissions_mapping():
    # wrap function factory, otherwise loop scope sticks the wrong way
    locals()[f"make_{name}_permissions"] = pytest.fixture(
        _make_permissions_factory(name, resource_type, levels, id_retriever)
    )

for name, resource_type, levels, id_retriever in _redash_permissions_mapping():
    # wrap function factory, otherwise loop scope sticks the wrong way
    locals()[f"make_{name}_permissions"] = pytest.fixture(
        _make_redash_permissions_factory(name, resource_type, levels, id_retriever)
    )


@pytest.fixture
def make_secret_scope(ws, make_random):
    def create(**kwargs):
        name = f"sdk-{make_random(4)}"
        ws.secrets.create_scope(name, **kwargs)
        return name

    yield from factory("secret scope", create, lambda scope: ws.secrets.delete_scope(scope))


@pytest.fixture
def make_secret_scope_acl(ws):
    def create(*, scope: str, principal: str, permission: workspace.AclPermission):
        ws.secrets.put_acl(scope, principal, permission)
        return scope, principal

    yield from factory("secret scope acl", create, lambda x: ws.secrets.delete_acl(x[0], x[1]))


@pytest.fixture
def make_notebook(ws, make_random):
    def create(
        *,
        path: str | Path | None = None,
        content: BinaryIO | None = None,
        language: Language = Language.PYTHON,
        format: ImportFormat = ImportFormat.SOURCE,  # pylint:  disable=redefined-builtin
        overwrite: bool = False,
    ) -> str:
        if path is None:
            path = f"/Users/{ws.current_user.me().user_name}/sdk-{make_random(4)}"
        elif isinstance(path, pathlib.Path):
            path = str(path)
        if content is None:
            content = io.BytesIO(b"print(1)")
        path = str(path)
        ws.workspace.upload(path, content, language=language, format=format, overwrite=overwrite)
        return path

    yield from factory("notebook", create, lambda x: ws.workspace.delete(x))


@pytest.fixture
def make_directory(ws, make_random):
    def create(*, path: str | None = None):
        if path is None:
            path = f"/Users/{ws.current_user.me().user_name}/sdk-{make_random(4)}"
        ws.workspace.mkdirs(path)
        return path

    yield from factory("directory", create, lambda x: ws.workspace.delete(x, recursive=True))


@pytest.fixture
def make_repo(ws, make_random):
    def create(*, url=None, provider=None, path=None, **kwargs):
        if path is None:
            path = f"/Repos/{ws.current_user.me().user_name}/sdk-{make_random(4)}"
        if url is None:
            url = "https://github.com/shreyas-goenka/empty-repo.git"
        if provider is None:
            provider = "github"
        return ws.repos.create(url, provider, path=path, **kwargs)

    yield from factory("repo", create, lambda x: ws.repos.delete(x.id))


@pytest.fixture
def make_user(ws, make_random):

    @retried(on=[ResourceConflict], timeout=timedelta(seconds=30))
    def create(**kwargs):
        user = ws.users.create(user_name=f"sdk-{make_random(8)}@example.com".lower(), **kwargs)
        return user

    yield from factory("workspace user", create, lambda item: ws.users.delete(item.id))


def _scim_values(ids: list[str]) -> list[iam.ComplexValue]:
    return [iam.ComplexValue(value=x) for x in ids]


def _make_group(name, cfg, interface, make_random):
    @retried(on=[ResourceConflict], timeout=timedelta(seconds=30))
    def create(
        *,
        members: list[str] | None = None,
        roles: list[str] | None = None,
        entitlements: list[str] | None = None,
        display_name: str | None = None,
        wait_for_provisioning: bool = False,
        **kwargs,
    ):
        kwargs["display_name"] = f"sdk-{make_random(4)}" if display_name is None else display_name
        if members is not None:
            kwargs["members"] = _scim_values(members)
        if roles is not None:
            kwargs["roles"] = _scim_values(roles)
        if entitlements is not None:
            kwargs["entitlements"] = _scim_values(entitlements)
        # TODO: REQUEST_LIMIT_EXCEEDED: GetUserPermissionsRequest RPC token bucket limit has been exceeded.
        group = interface.create(**kwargs)
        if cfg.is_account_client:
            logger.info(f"Account group {group.display_name}: {cfg.host}/users/groups/{group.id}/members")
        else:
            logger.info(f"Workspace group {group.display_name}: {cfg.host}#setting/accounts/groups/{group.id}")

        @retried(on=[NotFound], timeout=timedelta(minutes=2))
        def _wait_for_provisioning() -> None:
            interface.get(group.id)

        if wait_for_provisioning:
            _wait_for_provisioning()

        return group

    yield from factory(name, create, lambda item: interface.delete(item.id))


@pytest.fixture
def make_group(ws, make_random):
    yield from _make_group("workspace group", ws.config, ws.groups, make_random)


@pytest.fixture
def make_acc_group(acc, make_random):
    yield from _make_group("account group", acc.config, acc.groups, make_random)


@pytest.fixture
def migrated_group(acc, ws, make_group, make_acc_group):
    """Create a pair of groups in workspace and account. Assign account group to workspace."""
    ws_group = make_group()
    acc_group = make_acc_group()
    acc.workspace_assignment.update(ws.get_workspace_id(), acc_group.id, [iam.WorkspacePermission.USER])
    return MigratedGroup.partial_info(ws_group, acc_group)


@pytest.fixture
def make_cluster_policy(ws, make_random):
    def create(*, name: str | None = None, **kwargs):
        if name is None:
            name = f"sdk-{make_random(4)}"
        if "definition" not in kwargs:
            kwargs["definition"] = json.dumps(
                {
                    "spark_conf.spark.databricks.delta.preview.enabled": {"type": "fixed", "value": "true"},
                }
            )
        cluster_policy = ws.cluster_policies.create(name, **kwargs)
        logger.info(
            f"Cluster policy: {ws.config.host}#setting/clusters/cluster-policies/view/{cluster_policy.policy_id}"
        )
        return cluster_policy

    yield from factory("cluster policy", create, lambda item: ws.cluster_policies.delete(item.policy_id))


@pytest.fixture
def make_cluster(ws, make_random):
    def create(
        *,
        single_node: bool = False,
        cluster_name: str | None = None,
        spark_version: str | None = None,
        autotermination_minutes=10,
        **kwargs,
    ):
        if cluster_name is None:
            cluster_name = f"sdk-{make_random(4)}"
        if spark_version is None:
            spark_version = ws.clusters.select_spark_version(latest=True)
        if single_node:
            kwargs["num_workers"] = 0
            if "spark_conf" in kwargs:
                kwargs["spark_conf"] = kwargs["spark_conf"] | {
                    "spark.databricks.cluster.profile": "singleNode",
                    "spark.master": "local[*]",
                }
            else:
                kwargs["spark_conf"] = {"spark.databricks.cluster.profile": "singleNode", "spark.master": "local[*]"}
            kwargs["custom_tags"] = {"ResourceClass": "SingleNode"}
        if "instance_pool_id" not in kwargs:
            kwargs["node_type_id"] = ws.clusters.select_node_type(local_disk=True, min_memory_gb=16)

        if "custom_tags" not in kwargs:
            kwargs["custom_tags"] = {"RemoveAfter": get_test_purge_time()}
        else:
            kwargs["custom_tags"]["RemoveAfter"] = get_test_purge_time()
        return ws.clusters.create(
            cluster_name=cluster_name,
            spark_version=spark_version,
            autotermination_minutes=autotermination_minutes,
            **kwargs,
        )

    yield from factory("cluster", create, lambda item: ws.clusters.permanent_delete(item.cluster_id))


@pytest.fixture
def make_experiment(ws, make_random):
    def create(
        *,
        path: str | None = None,
        experiment_name: str | None = None,
        **kwargs,
    ):
        if path is None:
            path = f"/Users/{ws.current_user.me().user_name}/{make_random(4)}"
        if experiment_name is None:
            experiment_name = f"sdk-{make_random(4)}"

        try:
            ws.workspace.mkdirs(path)
        except DatabricksError:
            pass

        return ws.experiments.create_experiment(name=f"{path}/{experiment_name}", **kwargs)

    yield from factory("experiment", create, lambda item: ws.experiments.delete_experiment(item.experiment_id))


@pytest.fixture
def make_instance_pool(ws, make_random):
    def create(*, instance_pool_name=None, node_type_id=None, **kwargs):
        if instance_pool_name is None:
            instance_pool_name = f"sdk-{make_random(4)}"
        if node_type_id is None:
            node_type_id = ws.clusters.select_node_type(local_disk=True, min_memory_gb=16)
        return ws.instance_pools.create(
            instance_pool_name, node_type_id, custom_tags={"RemoveAfter": get_test_purge_time()}, **kwargs
        )

    yield from factory("instance pool", create, lambda item: ws.instance_pools.delete(item.instance_pool_id))


@pytest.fixture
def make_job(ws, make_random, make_notebook):
    def create(notebook_path: str | Path | None = None, **kwargs):
        task_spark_conf = None
        if "name" not in kwargs:
            kwargs["name"] = f"sdk-{make_random(4)}"
        if "spark_conf" in kwargs:
            task_spark_conf = kwargs["spark_conf"]
            kwargs.pop("spark_conf")
        libraries = None
        if "libraries" in kwargs:
            libraries = kwargs.pop("libraries")
        if isinstance(notebook_path, pathlib.Path):
            notebook_path = str(notebook_path)
        if not notebook_path:
            notebook_path = make_notebook()
        assert notebook_path is not None
        if "tasks" not in kwargs:
            kwargs["tasks"] = [
                jobs.Task(
                    task_key=make_random(4),
                    description=make_random(4),
                    new_cluster=compute.ClusterSpec(
                        num_workers=1,
                        node_type_id=ws.clusters.select_node_type(local_disk=True, min_memory_gb=16),
                        spark_version=ws.clusters.select_spark_version(latest=True),
                        spark_conf=task_spark_conf,
                    ),
                    notebook_task=jobs.NotebookTask(notebook_path=str(notebook_path)),
                    libraries=libraries,
                    timeout_seconds=0,
                )
            ]

        # add RemoveAfter tag for test job cleanup
        date_to_remove = get_test_purge_time()
        remove_after_tag = {"key": "RemoveAfter", "value": date_to_remove}

        if 'tags' not in kwargs:
            kwargs["tags"] = [remove_after_tag]
        else:
            kwargs["tags"].append(remove_after_tag)

        job = ws.jobs.create(**kwargs)
        logger.info(f"Job: {ws.config.host}#job/{job.job_id}")
        return job

    yield from factory("job", create, lambda item: ws.jobs.delete(item.job_id))


@pytest.fixture
def make_model(ws, make_random):
    def create(
        *,
        model_name: str | None = None,
        **kwargs,
    ):
        if model_name is None:
            model_name = f"sdk-{make_random(4)}"

        created_model = ws.model_registry.create_model(model_name, **kwargs)
        model = ws.model_registry.get_model(created_model.registered_model.name)
        return model.registered_model_databricks

    yield from factory("model", create, lambda item: ws.model_registry.delete_model(item.id))


@pytest.fixture
def make_pipeline(ws, make_random, make_notebook):
    def create(**kwargs) -> pipelines.CreatePipelineResponse:
        if "name" not in kwargs:
            kwargs["name"] = f"sdk-{make_random(4)}"
        if "libraries" not in kwargs:
            kwargs["libraries"] = [pipelines.PipelineLibrary(notebook=pipelines.NotebookLibrary(path=make_notebook()))]
        if "clusters" not in kwargs:
            kwargs["clusters"] = [
                pipelines.PipelineCluster(
                    node_type_id=ws.clusters.select_node_type(local_disk=True, min_memory_gb=16),
                    label="default",
                    num_workers=1,
                    custom_tags={
                        "cluster_type": "default",
                    },
                )
            ]
        return ws.pipelines.create(continuous=False, **kwargs)

    yield from factory("delta live table", create, lambda item: ws.pipelines.delete(item.pipeline_id))


@pytest.fixture
def make_warehouse(ws, make_random):
    def create(
        *,
        warehouse_name: str | None = None,
        warehouse_type: CreateWarehouseRequestWarehouseType | None = None,
        cluster_size: str | None = None,
        max_num_clusters: int = 1,
        enable_serverless_compute: bool = False,
        **kwargs,
    ):
        if warehouse_name is None:
            warehouse_name = f"sdk-{make_random(4)}"
        if warehouse_type is None:
            warehouse_type = CreateWarehouseRequestWarehouseType.PRO
        if cluster_size is None:
            cluster_size = "2X-Small"

        remove_after_tags = EndpointTags(custom_tags=[EndpointTagPair(key="RemoveAfter", value=get_test_purge_time())])
        return ws.warehouses.create(
            name=warehouse_name,
            cluster_size=cluster_size,
            warehouse_type=warehouse_type,
            max_num_clusters=max_num_clusters,
            enable_serverless_compute=enable_serverless_compute,
            tags=remove_after_tags,
            **kwargs,
        )

    yield from factory("warehouse", create, lambda item: ws.warehouses.delete(item.id))


def _is_in_debug() -> bool:
    return os.path.basename(sys.argv[0]) in {"_jb_pytest_runner.py", "testlauncher.py"}


@pytest.fixture
def debug_env_name():
    # Alternatively, we could use @pytest.mark.xxx, but
    # not sure how reusable it becomes then.
    #
    # we don't use scope=session, as monkeypatch.setenv
    # doesn't work on a session level
    return "UNKNOWN"


@pytest.fixture
def debug_env(monkeypatch, debug_env_name) -> MutableMapping[str, str]:
    if not _is_in_debug():
        return os.environ
    conf_file = pathlib.Path.home() / ".databricks/debug-env.json"
    if not conf_file.exists():
        return os.environ
    with conf_file.open("r") as f:
        conf = json.load(f)
        if debug_env_name not in conf:
            sys.stderr.write(f"""{debug_env_name} not found in ~/.databricks/debug-env.json""")
            msg = f"{debug_env_name} not found in ~/.databricks/debug-env.json"
            raise KeyError(msg)
        for k, v in conf[debug_env_name].items():
            monkeypatch.setenv(k, v)
    return os.environ


@pytest.fixture
def env_or_skip(debug_env) -> Callable[[str], str]:
    skip = pytest.skip
    if _is_in_debug():
        skip = pytest.fail  # type: ignore[assignment]

    def inner(var: str) -> str:
        if var not in debug_env:
            skip(f"Environment variable {var} is missing")
        return debug_env[var]

    return inner


@pytest.fixture
def sql_backend(ws, env_or_skip) -> StatementExecutionBackend:
    warehouse_id = env_or_skip("TEST_DEFAULT_WAREHOUSE_ID")
    return StatementExecutionBackend(ws, warehouse_id)


@pytest.fixture
def inventory_schema(make_schema):
    return make_schema(catalog_name="hive_metastore").name


@pytest.fixture
def make_catalog(ws, sql_backend, make_random) -> Generator[Callable[..., CatalogInfo], None, None]:
    def create() -> CatalogInfo:
        name = f"ucx_C{make_random(4)}".lower()
        sql_backend.execute(f"CREATE CATALOG {name}")
        catalog_info = ws.catalogs.get(name)
        return catalog_info

    yield from factory(
        "catalog",
        create,
        lambda catalog_info: ws.catalogs.delete(catalog_info.full_name, force=True),
    )


@pytest.fixture
def make_schema(ws, sql_backend, make_random) -> Generator[Callable[..., SchemaInfo], None, None]:
    def create(*, catalog_name: str = "hive_metastore", name: str | None = None) -> SchemaInfo:
        if name is None:
            name = f"ucx_S{make_random(4)}".lower()
        full_name = f"{catalog_name}.{name}".lower()
        sql_backend.execute(f"CREATE SCHEMA {full_name} WITH DBPROPERTIES (RemoveAfter={get_test_purge_time()})")
        schema_info = SchemaInfo(catalog_name=catalog_name, name=name, full_name=full_name)
        logger.info(
            f"Schema {schema_info.full_name}: "
            f"{ws.config.host}/explore/data/{schema_info.catalog_name}/{schema_info.name}"
        )
        return schema_info

    def remove(schema_info: SchemaInfo):
        try:
            sql_backend.execute(f"DROP SCHEMA IF EXISTS {schema_info.full_name} CASCADE")
        except RuntimeError as e:
            if "SCHEMA_NOT_FOUND" in str(e):
                logger.warning("Schema was already dropped while executing the test", exc_info=e)
            else:
                raise e

    yield from factory("schema", create, remove)


@pytest.fixture
# pylint: disable-next=too-many-statements
def make_table(ws, sql_backend, make_schema, make_random) -> Generator[Callable[..., TableInfo], None, None]:
    def create(  # pylint: disable=too-many-locals,too-many-arguments,too-many-statements
        *,
        catalog_name="hive_metastore",
        schema_name: str | None = None,
        name: str | None = None,
        ctas: str | None = None,
        non_delta: bool = False,
        external: bool = False,
        external_csv: str | None = None,
        external_delta: str | None = None,
        view: bool = False,
        tbl_properties: dict[str, str] | None = None,
        hiveserde_ddl: str | None = None,
        storage_override: str | None = None,
    ) -> TableInfo:
        if schema_name is None:
            schema = make_schema(catalog_name=catalog_name)
            catalog_name = schema.catalog_name
            schema_name = schema.name
        if name is None:
            name = f"ucx_T{make_random(4)}".lower()
        table_type: TableType | None = None
        data_source_format = None
        storage_location = None
        view_text = None
        full_name = f"{catalog_name}.{schema_name}.{name}".lower()
        ddl = f'CREATE {"VIEW" if view else "TABLE"} {full_name}'
        if view:
            table_type = TableType.VIEW
            view_text = ctas
        if ctas is not None:
            # temporary (if not view)
            ddl = f"{ddl} AS {ctas}"
        elif non_delta:
            table_type = TableType.EXTERNAL  # pylint: disable=redefined-variable-type
            data_source_format = DataSourceFormat.JSON
            storage_location = f"dbfs:/tmp/ucx_test_{make_random(4)}"
            # Modified, otherwise it will identify the table as a DB Dataset
            ddl = (
                f"{ddl} USING json location '{storage_location}' as SELECT * FROM "
                f"JSON.`dbfs:/databricks-datasets/iot-stream/data-device`"
            )
        elif external_csv is not None:
            table_type = TableType.EXTERNAL
            data_source_format = DataSourceFormat.CSV
            storage_location = external_csv
            ddl = f"{ddl} USING CSV OPTIONS (header=true) LOCATION '{storage_location}'"
        elif external_delta is not None:
            table_type = TableType.EXTERNAL
            data_source_format = DataSourceFormat.DELTA
            storage_location = external_delta
            ddl = f"{ddl} (id string) LOCATION '{storage_location}'"
        elif external:
            # external table
            table_type = TableType.EXTERNAL
            data_source_format = DataSourceFormat.DELTASHARING
            url = "s3a://databricks-datasets-oregon/delta-sharing/share/open-datasets.share"
            storage_location = f"{url}#delta_sharing.default.lending_club"
            ddl = f"{ddl} USING deltaSharing LOCATION '{storage_location}'"
        else:
            # managed table
            table_type = TableType.MANAGED
            data_source_format = DataSourceFormat.DELTA
            storage_location = f"dbfs:/user/hive/warehouse/{schema_name}/{name}"
            ddl = f"{ddl} (id INT, value STRING)"
        if tbl_properties:
            tbl_properties.update({"RemoveAfter": get_test_purge_time()})
        else:
            tbl_properties = {"RemoveAfter": get_test_purge_time()}

        str_properties = ",".join([f" '{k}' = '{v}' " for k, v in tbl_properties.items()])

        # table properties fails with CTAS statements
        alter_table_tbl_properties = ""
        if ctas or non_delta:
            alter_table_tbl_properties = (
                f'ALTER {"VIEW" if view else "TABLE"} {full_name} SET TBLPROPERTIES ({str_properties})'
            )
        else:
            ddl = f"{ddl} TBLPROPERTIES ({str_properties})"

        if hiveserde_ddl:
            ddl = hiveserde_ddl
            data_source_format = None
            table_type = TableType.EXTERNAL
            storage_location = storage_override

        sql_backend.execute(ddl)

        # CTAS AND NON_DELTA does not support TBLPROPERTIES
        if ctas or non_delta:
            sql_backend.execute(alter_table_tbl_properties)

        table_info = TableInfo(
            catalog_name=catalog_name,
            schema_name=schema_name,
            name=name,
            full_name=full_name,
            properties=tbl_properties,
            storage_location=storage_location,
            table_type=table_type,
            view_definition=view_text,
            data_source_format=data_source_format,
        )
        logger.info(
            f"Table {table_info.full_name}: "
            f"{ws.config.host}/explore/data/{table_info.catalog_name}/{table_info.schema_name}/{table_info.name}"
        )
        return table_info

    def remove(table_info: TableInfo):
        try:
            sql_backend.execute(f"DROP TABLE IF EXISTS {table_info.full_name}")
        except RuntimeError as e:
            if "Cannot drop a view" in str(e):
                sql_backend.execute(f"DROP VIEW IF EXISTS {table_info.full_name}")
            elif "SCHEMA_NOT_FOUND" in str(e):
                logger.warning("Schema was already dropped while executing the test", exc_info=e)
            else:
                raise e

    yield from factory("table", create, remove)


@pytest.fixture
def make_udf(
    ws,
    env_or_skip,
    sql_backend,
    make_schema,
    make_random,
) -> Generator[Callable[..., FunctionInfo], None, None]:
    def create(
        *,
        catalog_name="hive_metastore",
        schema_name: str | None = None,
        name: str | None = None,
        hive_udf: bool = False,
    ) -> FunctionInfo:
        if schema_name is None:
            schema = make_schema(catalog_name=catalog_name)
            catalog_name = schema.catalog_name
            schema_name = schema.name

        if name is None:
            name = f"ucx_T{make_random(4)}".lower()

        full_name = f"{catalog_name}.{schema_name}.{name}".lower()
        if hive_udf:
            cmd_exec = CommandExecutor(
                ws.clusters,
                ws.command_execution,
                lambda: env_or_skip("TEST_DEFAULT_CLUSTER_ID"),
                language=compute.Language.SQL,
            )
            ddl = f"CREATE FUNCTION {full_name} AS 'org.apache.hadoop.hive.ql.udf.generic.GenericUDFAbs';"
            cmd_exec.run(ddl)
        else:
            ddl = f"CREATE FUNCTION {full_name}(x INT) RETURNS FLOAT CONTAINS SQL DETERMINISTIC RETURN 0;"
            sql_backend.execute(ddl)
        udf_info = FunctionInfo(
            catalog_name=catalog_name,
            schema_name=schema_name,
            name=name,
            full_name=full_name,
        )

        logger.info(f"Function {udf_info.full_name} crated")
        return udf_info

    def remove(udf_info: FunctionInfo):
        try:
            sql_backend.execute(f"DROP FUNCTION IF EXISTS {udf_info.full_name}")
        except NotFound as e:
            if "SCHEMA_NOT_FOUND" in str(e):
                logger.warning("Schema was already dropped while executing the test", exc_info=e)
            else:
                raise e

    yield from factory("table", create, remove)


@pytest.fixture
def make_query(ws, make_table, make_random) -> Generator[Query, None, None]:
    def create() -> Query:
        table = make_table()
        query_name = f"ucx_query_Q{make_random(4)}"
        query = ws.queries.create(
            name=query_name,
            description="TEST QUERY FOR UCX",
            query=f"SELECT * FROM {table.schema_name}.{table.name}",
            tags=["original_query_tag"],
        )
        logger.info(f"Query Created {query_name}: {ws.config.host}/sql/editor/{query.id}")
        return query

    def remove(query: Query):
        try:
            ws.queries.delete(query_id=query.id)
        except RuntimeError as e:
            logger.info(f"Can't drop query {e}")

    yield from factory("query", create, remove)


@pytest.fixture
def make_storage_credential(ws):
    def create(
        *,
        credential_name: str,
        application_id: str = "",
        client_secret: str = "",
        directory_id: str = "",
        aws_iam_role_arn: str = "",
        read_only=False,
    ) -> StorageCredentialInfo:
        if aws_iam_role_arn != "":
            storage_credential = ws.storage_credentials.create(
                credential_name, aws_iam_role=AwsIamRoleRequest(role_arn=aws_iam_role_arn), read_only=read_only
            )
        else:
            azure_service_principal = AzureServicePrincipal(directory_id, application_id, client_secret)
            storage_credential = ws.storage_credentials.create(
                credential_name, azure_service_principal=azure_service_principal, read_only=read_only
            )
        return storage_credential

    def remove(storage_credential: StorageCredentialInfo):
        ws.storage_credentials.delete(storage_credential.name, force=True)

    yield from factory("storage_credential", create, remove)


@pytest.fixture
def make_serving_endpoint(ws, make_random, make_model):
    def create() -> Wait[ServingEndpointDetailed]:
        endpoint_name = make_random(4)
        model = make_model()
        endpoint = ws.serving_endpoints.create(
            endpoint_name,
            EndpointCoreConfigInput(
                served_models=[
                    ServedModelInput(model.name, "1", ServedModelInputWorkloadSize.SMALL, scale_to_zero_enabled=True)
                ]
            ),
        )
        return endpoint

    def remove(endpoint_name: str):
        try:
            ws.serving_endpoints.delete(endpoint_name)
        except RuntimeError as e:
            logger.info(f"Can't remove endpoint {e}")

    yield from factory("Serving endpoint", create, remove)


@pytest.fixture
def make_feature_table(ws, make_random):
    def create():
        feature_table_name = make_random(6) + "." + make_random(6)
        table = ws.api_client.do(
            "POST",
            "/api/2.0/feature-store/feature-tables/create",
            body={"name": feature_table_name, "primary_keys": [{"name": "pk", "data_type": "string"}]},
        )
        return table['feature_table']

    def remove(table: dict):
        try:
            ws.api_client.do("DELETE", "/api/2.0/feature-store/feature-tables/delete", body={"name": table["name"]})
        except RuntimeError as e:
            logger.info(f"Can't remove feature table {e}")

    yield from factory("Feature table", create, remove)


@pytest.fixture
def make_dbfs_data_copy(ws, make_cluster, env_or_skip):
    _ = make_cluster  # Need cluster to copy data
    if ws.config.is_aws:
        cmd_exec = CommandExecutor(ws.clusters, ws.command_execution, lambda: env_or_skip("TEST_WILDCARD_CLUSTER_ID"))

    def create(*, src_path: str, dst_path: str, wait_for_provisioning=True):
        @retried(on=[NotFound], timeout=timedelta(minutes=2))
        def _wait_for_provisioning(path) -> None:
            if not ws.dbfs.exists(path):
                raise NotFound(f"Location not found: {path}")

        if ws.config.is_aws:
            cmd_exec.run(f"dbutils.fs.cp('{src_path}', '{dst_path}', recurse=True)")
        else:
            ws.dbfs.copy(src_path, dst_path, recursive=True)
            if wait_for_provisioning:
                _wait_for_provisioning(dst_path)
        return dst_path

    def remove(dst_path: str):
        if ws.config.is_aws:
            cmd_exec.run(f"dbutils.fs.rm('{dst_path}', recurse=True)")
        else:
            ws.dbfs.delete(dst_path, recursive=True)

    yield from factory("make_dbfs_data_copy", create, remove)


@pytest.fixture
def make_mounted_location(make_random, make_dbfs_data_copy, env_or_skip):
    """Make a copy of source data to a new location

    Use the fixture to avoid overlapping UC table path that will fail other external table migration tests.

    Note:
        This fixture is different to the other `make_` fixtures as it does not return a `Callable` to make the mounted
        location; the mounted location is made with fixture setup already.
    """
    existing_mounted_location = f'dbfs:/mnt/{env_or_skip("TEST_MOUNT_NAME")}/a/b/c'
    new_mounted_location = f'dbfs:/mnt/{env_or_skip("TEST_MOUNT_NAME")}/a/b/{make_random(4)}'
    make_dbfs_data_copy(src_path=existing_mounted_location, dst_path=new_mounted_location)
    return new_mounted_location


@pytest.fixture
def make_storage_dir(ws, env_or_skip):
    if ws.config.is_aws:
        cmd_exec = CommandExecutor(ws.clusters, ws.command_execution, lambda: env_or_skip("TEST_WILDCARD_CLUSTER_ID"))

    def create(*, path: str):
        if ws.config.is_aws:
            cmd_exec.run(f"dbutils.fs.mkdirs('{path}')")
        else:
            ws.dbfs.mkdirs(path)
        return path

    def remove(path: str):
        if ws.config.is_aws:
            cmd_exec.run(f"dbutils.fs.rm('{path}', recurse=True)")
        else:
            ws.dbfs.delete(path, recursive=True)

    yield from factory("make_storage_dir", create, remove)


@pytest.fixture
def make_dashboard(ws: WorkspaceClient, make_random: Callable[[int], str], make_query):
    def create() -> Dashboard:
        query = make_query()
        viz = ws.query_visualizations.create(
            type="table",
            query_id=query.id,
            options={
                "itemsPerPage": 1,
                "condensed": True,
                "withRowNumber": False,
                "version": 2,
                "columns": [
                    {"name": "id", "title": "id", "allowSearch": True},
                ],
            },
        )

        dashboard_name = f"ucx_D{make_random(4)}"
        dashboard = ws.dashboards.create(name=dashboard_name, tags=["original_dashboard_tag"])
        assert dashboard.id is not None
        ws.dashboard_widgets.create(
            dashboard_id=dashboard.id,
            visualization_id=viz.id,
            width=1,
            options=WidgetOptions(
                title="",
                position=WidgetPosition(
                    col=0,
                    row=0,
                    size_x=3,
                    size_y=3,
                ),
            ),
        )
        logger.info(f"Dashboard Created {dashboard_name}: {ws.config.host}/sql/dashboards/{dashboard.id}")
        return dashboard

    def remove(dashboard: Dashboard) -> None:
        try:
            assert dashboard.id is not None
            ws.dashboards.delete(dashboard_id=dashboard.id)
        except RuntimeError as e:
            logger.info(f"Can't delete dashboard {e}")

    yield from factory("dashboard", create, remove)


def get_test_purge_time() -> str:
    return (datetime.utcnow() + TEST_JOBS_PURGE_TIMEOUT).strftime("%Y%m%d%H")



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from astroid import Attribute, Call, NodeNG  # type: ignore

from databricks.labs.ucx.source_code.base import (
    Advice,
    PythonLinter,
)
from databricks.labs.ucx.source_code.linters.python_ast import Tree


@dataclass
class Position:
    # Lines and columns are both 0-based.
    line: int
    character: int


@dataclass
class Range:
    start: Position
    end: Position


@dataclass
class NoFormatPythonMatcher:
    """Matches Python AST nodes where tables are created with implicit format.
    It performs only the matching, while linting / fixing are separated concerns.
    """

    method_name: str
    min_args: int
    max_args: int
    # some method_names accept 'format' as a direct (optional) argument:
    format_arg_index: int | None = None
    format_arg_name: str | None = None

    def get_advice_span(self, node: NodeNG) -> Range | None:
        # Check 1: check Call:
        if not isinstance(node, Call):
            return None

        # Check 2: check presence of the table-creating method call:
        if not isinstance(node.func, Attribute) or node.func.attrname != self.method_name:
            return None
        call_args_count = Tree.args_count(node)
        if call_args_count < self.min_args or call_args_count > self.max_args:
            return None

        # Check 3: check presence of the format specifier:
        #   Option A: format specifier may be given as a direct parameter to the table-creating call
        #   >>> df.saveToTable("c.db.table", format="csv")
        format_arg = Tree.get_arg(node, self.format_arg_index, self.format_arg_name)
        if format_arg is not None and not Tree.is_none(format_arg):
            # i.e., found an explicit "format" argument, and its value is not None.
            return None
        #   Option B. format specifier may be a separate ".format(...)" call in this callchain
        #   >>> df.format("csv").saveToTable("c.db.table")
        format_call = Tree.extract_call_by_name(node, "format")
        if format_call is not None:
            # i.e., found an explicit ".format(...)" call in this chain.
            return None

        # Finally: matched the need for advice, so return the corresponding source range.
        # Note that astroid line numbers are 1-based.
        return Range(
            Position(node.lineno - 1, node.col_offset),
            Position((node.end_lineno or 1) - 1, node.end_col_offset or 0),
        )


class NoFormatPythonLinter:
    """Python linting for table-creation with implicit format"""

    def __init__(self, matchers: list[NoFormatPythonMatcher]):
        self._matchers = matchers

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        for matcher in self._matchers:
            span = matcher.get_advice_span(node)
            if span is not None:
                yield Advice(
                    code="table-migrate",
                    message="The default format changed in Databricks Runtime 8.0, from Parquet to Delta",
                    start_line=span.start.line,
                    start_col=span.start.character,
                    end_line=span.end.line,
                    end_col=span.end.character,
                )


class DBRv8d0Linter(PythonLinter):
    """Performs Python linting for backwards incompatible changes in DBR version 8.0.
    Specifically, it yields advice for table-creation with implicit format.
    """

    # https://docs.databricks.com/en/archive/runtime-release-notes/8.0.html#delta-is-now-the-default-format-when-a-format-is-not-specified

    def __init__(self, dbr_version: tuple[int, int] | None):
        version_cutoff = (8, 0)
        self._skip_dbr = dbr_version is not None and dbr_version >= version_cutoff

        self._linter = NoFormatPythonLinter(
            [
                NoFormatPythonMatcher("writeTo", 1, 1),
                NoFormatPythonMatcher("table", 1, 1),
                NoFormatPythonMatcher("saveAsTable", 1, 4, 2, "format"),
            ]
        )

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        if self._skip_dbr:
            return
        for node in tree.walk():
            yield from self._linter.lint(node)



import dataclasses
import logging
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path

from databricks.labs.blueprint.installer import InstallState
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError, NotFound
from databricks.sdk.service import workspace
from databricks.sdk.service.sql import (
    AccessControl,
    ObjectTypePlural,
    PermissionLevel,
    RunAsRole,
    WidgetOptions,
    WidgetPosition,
)

logger = logging.getLogger(__name__)

# pylint: disable=invalid-name


@dataclass
class SimpleQuery:
    dashboard_ref: str
    name: str
    widget: dict[str, str]
    viz: dict[str, str]
    query: str | None = None
    text: str | None = None

    @property
    def key(self):
        return f"{self.dashboard_ref}_{self.name}"

    @property
    def viz_type(self) -> str:
        return self.viz.get("type", "UNKNOWN")

    @property
    def viz_args(self) -> dict:
        return {k: v for k, v in self.viz.items() if k not in ["type"]}


@dataclass
class VizColumn:  # pylint: disable=too-many-instance-attributes)
    name: str
    title: str
    type: str = "string"
    imageUrlTemplate: str = "{{ @ }}"  # noqa: N815
    imageTitleTemplate: str = "{{ @ }}"  # noqa: N815
    linkUrlTemplate: str = "{{ @ }}"  # noqa: N815
    linkTextTemplate: str = "{{ @ }}"  # noqa: N815
    linkTitleTemplate: str = "{{ @ }}"  # noqa: N815
    linkOpenInNewTab: bool = True  # noqa: N815
    displayAs: str = "string"  # noqa: N815
    visible: bool = True
    order: int = 100000
    allowSearch: bool = False  # noqa: N815
    alignContent: str = "left"  # noqa: N815
    allowHTML: bool = False  # noqa: N815
    highlightLinks: bool = False  # noqa: N815
    useMonospaceFont: bool = False  # noqa: N815
    preserveWhitespace: bool = False  # noqa: N815

    def as_dict(self):
        return dataclasses.asdict(self)


class DashboardFromFiles:
    def __init__(
        self,
        ws: WorkspaceClient,
        state: InstallState,
        local_folder: Path,
        remote_folder: str,
        name_prefix: str,
        query_text_callback: Callable[[str], str] | None = None,
        warehouse_id: str | None = None,
    ):
        self._ws = ws
        self._local_folder = local_folder
        self._remote_folder = remote_folder
        self._name_prefix = name_prefix
        self._query_text_callback = query_text_callback
        self._warehouse_id = warehouse_id
        self._state = state
        self._pos = 0

    def dashboard_link(self, dashboard_ref: str):
        dashboard_id = self._state.dashboards[dashboard_ref]
        return f"{self._ws.config.host}/sql/dashboards/{dashboard_id}"

    def create_dashboards(self) -> dict[str, str]:
        queries_per_dashboard = {}
        # Iterate over dashboards for each step, represented as first-level folders
        step_folders = [f for f in self._local_folder.glob("*") if f.is_dir()]
        for step_folder in step_folders:
            logger.debug(f"Reading step folder {step_folder}...")
            dashboard_folders = [f for f in step_folder.glob("*") if f.is_dir()]
            # Create separate dashboards per step, represented as second-level folders
            for dashboard_folder in dashboard_folders:
                logger.debug(f"Reading dashboard folder {dashboard_folder}...")
                main_name = step_folder.stem.title()
                sub_name = dashboard_folder.stem.title()
                dashboard_name = f"{self._name_prefix} {main_name} ({sub_name})"
                dashboard_ref = f"{step_folder.stem}_{dashboard_folder.stem}".lower()
                logger.info(f"Creating dashboard {dashboard_name}...")
                desired_queries = self._desired_queries(dashboard_folder, dashboard_ref)
                parent_folder_id = self._installed_query_state()
                data_source_id = self._dashboard_data_source()
                self._install_dashboard(dashboard_name, parent_folder_id, dashboard_ref)
                for query in desired_queries:
                    self._install_query(query, dashboard_name, data_source_id, parent_folder_id)
                    self._install_viz(query)
                    self._install_widget(query, dashboard_ref)
                queries_per_dashboard[dashboard_ref] = desired_queries
        self._store_query_state(queries_per_dashboard)
        return self._state.dashboards

    def validate(self):
        step_folders = [f for f in self._local_folder.glob("*") if f.is_dir()]
        for step_folder in step_folders:
            logger.info(f"Reading step folder {step_folder}...")
            dashboard_folders = [f for f in step_folder.glob("*") if f.is_dir()]
            # Create separate dashboards per step, represented as second-level folders
            for dashboard_folder in dashboard_folders:
                self._validate_folder(dashboard_folder, step_folder)

    def _validate_folder(self, dashboard_folder, step_folder):
        dashboard_ref = f"{step_folder.stem}_{dashboard_folder.stem}".lower()
        for query in self._desired_queries(dashboard_folder, dashboard_ref):
            if query.text:
                continue
            try:
                self._get_viz_options(query)
                self._get_widget_options(query)
            except Exception as err:
                msg = f"Error in {query.name}: {err}"
                raise AssertionError(msg) from err

    def _install_widget(self, query: SimpleQuery, dashboard_ref: str):
        dashboard_id = self._state.dashboards[dashboard_ref]
        widget_options = self._get_widget_options(query)
        # widgets are cleaned up every dashboard redeploy
        if query.query:
            widget = self._ws.dashboard_widgets.create(
                dashboard_id, widget_options, 1, visualization_id=self._state.viz[query.key]
            )
        elif query.text:
            text = query.text[query.text.index("\n") + 1 :]
            widget = self._ws.dashboard_widgets.create(dashboard_id, widget_options, 1, text=text)
        else:
            raise ValueError("Query or Text should be set")
        assert widget.id is not None
        self._state.widgets[query.key] = widget.id

    def _get_widget_options(self, query: SimpleQuery):
        self._pos += 1
        widget_options = WidgetOptions(
            title=query.widget.get("title", ""),
            description=query.widget.get("description", None),
            position=WidgetPosition(
                col=int(query.widget.get("col", 0)),
                row=int(query.widget.get("row", self._pos)),
                size_x=int(query.widget.get("size_x", 3)),
                size_y=int(query.widget.get("size_y", 3)),
            ),
        )
        return widget_options

    def _remote_folder_object(self) -> workspace.ObjectInfo:
        try:
            return self._ws.workspace.get_status(self._remote_folder)
        except NotFound:
            self._ws.workspace.mkdirs(self._remote_folder)
            return self._remote_folder_object()

    def _installed_query_state(self):
        object_info = self._remote_folder_object()
        parent = f"folders/{object_info.object_id}"
        return parent

    def _store_query_state(self, queries_per_dashboard: dict[str, list[SimpleQuery]]):  # pylint: disable=too-complex
        query_refs = set()
        dashboard_refs = queries_per_dashboard.keys()
        for queries in queries_per_dashboard.values():
            for query in queries:
                query_refs.add(query.key)

        def silent_destroy(fn, object_id):
            try:
                fn(object_id)
            except DatabricksError as err:
                logger.info(f"Failed to delete {object_id} --- {err.error_code}")

        for ref, object_id in self._state.dashboards.items():
            if ref in dashboard_refs:
                continue
            silent_destroy(self._ws.dashboards.delete, object_id)

        for ref, object_id in self._state.queries.items():
            if ref in query_refs:
                continue
            silent_destroy(self._ws.queries.delete, object_id)

        for ref, object_id in self._state.viz.items():
            if ref in query_refs:
                continue
            silent_destroy(self._ws.query_visualizations.delete, object_id)

        for ref, object_id in self._state.widgets.items():
            if ref in query_refs:
                continue
            silent_destroy(self._ws.dashboard_widgets.delete, object_id)

        self._state.save()

    def _install_dashboard(self, dashboard_name: str, parent_folder_id: str, dashboard_ref: str):
        if dashboard_ref in self._state.dashboards:
            dashboard = self._ws.dashboards.get(self._state.dashboards[dashboard_ref])
            assert dashboard.widgets is not None
            for widget in dashboard.widgets:
                assert widget.id is not None
                try:
                    self._ws.dashboard_widgets.delete(widget.id)
                except TypeError:
                    pass
            return
        dashboard = self._ws.dashboards.create(dashboard_name, run_as_role=RunAsRole.VIEWER, parent=parent_folder_id)
        assert dashboard.id is not None
        self._ws.dbsql_permissions.set(
            ObjectTypePlural.DASHBOARDS,
            dashboard.id,
            access_control_list=[AccessControl(group_name="users", permission_level=PermissionLevel.CAN_VIEW)],
        )
        self._state.dashboards[dashboard_ref] = dashboard.id

    def _desired_queries(self, local_folder: Path, dashboard_ref: str) -> list[SimpleQuery]:
        desired_queries = []
        for f in local_folder.glob("*.sql"):
            text = f.read_text("utf8")
            if self._query_text_callback is not None:
                text = self._query_text_callback(text)
            desired_queries.append(
                SimpleQuery(
                    dashboard_ref=dashboard_ref,
                    name=f.name,
                    query=text,
                    viz=self._parse_magic_comment(f, "-- viz ", text),
                    widget=self._parse_magic_comment(f, "-- widget ", text),
                )
            )
        for f in local_folder.glob("*.md"):
            if f.name == "README.md":
                continue
            text = f.read_text("utf8")
            desired_queries.append(
                SimpleQuery(
                    dashboard_ref=dashboard_ref,
                    name=f.name,
                    text=text,
                    widget=self._parse_magic_comment(f, "-- widget ", text),
                    viz={},
                )
            )
        return desired_queries

    def _install_viz(self, query: SimpleQuery):
        if query.text:
            logger.debug(f"Skipping viz {query.name} because it's a text widget")
            return None
        viz_args = self._get_viz_options(query)
        if query.key in self._state.viz:
            return self._ws.query_visualizations.update(self._state.viz[query.key], **viz_args)
        viz = self._ws.query_visualizations.create(self._state.queries[query.key], **viz_args)
        assert viz.id is not None
        self._state.viz[query.key] = viz.id
        return None

    def _get_viz_options(self, query: SimpleQuery):
        viz_types: dict[str, Callable[..., dict]] = {"table": self._table_viz_args, "counter": self._counter_viz_args}
        if query.viz_type not in viz_types:
            msg = f"{query.query}: unknown viz type: {query.viz_type}"
            raise SyntaxError(msg)
        viz_args = viz_types[query.viz_type](**query.viz_args)
        return viz_args

    def _install_query(self, query: SimpleQuery, dashboard_name: str, data_source_id: str, parent: str):
        if query.text:
            logger.debug(f"Skipping query {query.name} because it's a text widget")
            return None
        query_meta = {
            "data_source_id": data_source_id,
            "name": f"{dashboard_name} - {query.name}",
            "query": query.query,
        }
        if query.key in self._state.queries:
            return self._ws.queries.update(self._state.queries[query.key], **query_meta, tags=None, run_as_role=None)

        deployed_query = self._ws.queries.create(parent=parent, run_as_role=RunAsRole.VIEWER, **query_meta, tags=None)
        assert deployed_query.id is not None
        self._ws.dbsql_permissions.set(
            ObjectTypePlural.QUERIES,
            deployed_query.id,
            access_control_list=[AccessControl(group_name="users", permission_level=PermissionLevel.CAN_RUN)],
        )
        self._state.queries[query.key] = deployed_query.id
        return None

    @staticmethod
    def _table_viz_args(
        name: str,
        columns: str,
        *,
        items_per_page: int = 25,
        condensed=True,
        with_row_number=False,
        description: str | None = None,
        search_by: str | None = None,
    ) -> dict:
        if search_by is not None:
            search_by_columns = set(search_by.split(","))
        else:
            search_by_columns = set()
        return {
            "type": "TABLE",
            "name": name,
            "description": description,
            "options": {
                "itemsPerPage": items_per_page,
                "condensed": condensed,
                "withRowNumber": with_row_number,
                "version": 2,
                "columns": [
                    VizColumn(name=x, title=x, allowSearch=x in search_by_columns).as_dict() for x in columns.split(",")
                ],
            },
        }

    @staticmethod
    def _counter_viz_args(  # pylint: disable=too-many-arguments
        name: str,
        value_column: str,
        *,
        description: str | None = None,
        counter_label: str | None = None,
        value_row_number: int = 1,
        target_row_number: int = 1,
        string_decimal: int = 0,
        string_decimal_char: str = ".",
        string_thousand_separator: str = ",",
        tooltip_format: str = "0,0.000",
        count_row: bool = False,
    ) -> dict:
        return {
            "type": "COUNTER",
            "name": name,
            "description": description,
            "options": {
                "counterLabel": counter_label,
                "counterColName": value_column,
                "rowNumber": value_row_number,
                "targetRowNumber": target_row_number,
                "stringDecimal": string_decimal,
                "stringDecChar": string_decimal_char,
                "stringThouSep": string_thousand_separator,
                "tooltipFormat": tooltip_format,
                "countRow": count_row,
            },
        }

    @staticmethod
    def _parse_magic_comment(f, magic_comment, text):
        viz_comment = next(_ for _ in text.splitlines() if _.startswith(magic_comment))
        if not viz_comment:
            msg = f'{f}: cannot find "{magic_comment}" magic comment'
            raise SyntaxError(msg)
        return dict(_.split("=") for _ in viz_comment.replace(magic_comment, "").split(", "))

    def _dashboard_data_source(self) -> str:
        data_sources = {_.warehouse_id: _.id for _ in self._ws.data_sources.list()}
        warehouses = list(self._ws.warehouses.list())
        warehouse_id = self._warehouse_id
        if not warehouse_id and not warehouses:
            msg = "need either configured warehouse_id or an existing SQL warehouse"
            raise ValueError(msg)
        if not warehouse_id:
            warehouse_id = warehouses[0].id
        data_source_id = data_sources[warehouse_id]
        return data_source_id



from __future__ import annotations

import abc
import logging
from pathlib import Path

from databricks.sdk.errors import NotFound

from databricks.labs.ucx.source_code.graph import (
    BaseNotebookResolver,
    Dependency,
    DependencyLoader,
    MaybeDependency,
    SourceContainer,
)
from databricks.labs.ucx.source_code.notebooks.cells import CellLanguage
from databricks.labs.ucx.source_code.notebooks.sources import Notebook, SUPPORTED_EXTENSION_LANGUAGES
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.getLogger(__name__)


class NotebookResolver(BaseNotebookResolver):

    def __init__(self, notebook_loader: NotebookLoader):
        super().__init__()
        self._notebook_loader = notebook_loader

    def resolve_notebook(self, path_lookup: PathLookup, path: Path) -> MaybeDependency:
        absolute_path = self._notebook_loader.resolve(path_lookup, path)
        if not absolute_path:
            return self._fail('notebook-not-found', f"Notebook not found: {path.as_posix()}")
        dependency = Dependency(self._notebook_loader, absolute_path)
        return MaybeDependency(dependency, [])


class NotebookLoader(DependencyLoader, abc.ABC):
    def resolve(self, path_lookup: PathLookup, path: Path) -> Path | None:
        """When exported through Git, notebooks are saved with a .py extension. If the path is a notebook, return the
        path to the notebook. If the path is a Python file, return the path to the Python file. If the path is neither,
        return None."""
        for candidate in (path, self._adjust_path(path)):
            absolute_path = path_lookup.resolve(candidate)
            if not absolute_path:
                continue
            return absolute_path
        return None

    def load_dependency(self, path_lookup: PathLookup, dependency: Dependency) -> SourceContainer | None:
        absolute_path = self.resolve(path_lookup, dependency.path)
        if not absolute_path:
            return None
        try:
            content = absolute_path.read_text("utf-8")
        except NotFound:
            logger.warning(f"Path not found trying to read notebook from workspace: {absolute_path}")
            return None
        except PermissionError:
            logger.warning(
                f"Permission error while reading notebook from workspace: {absolute_path}",
                exc_info=True,
            )
            return None
        language = self._detect_language(absolute_path, content)
        if not language:
            logger.warning(f"Could not detect language for {absolute_path}")
            return None
        return Notebook.parse(absolute_path, content, language)

    @staticmethod
    def _detect_language(path: Path, content: str):
        language = SUPPORTED_EXTENSION_LANGUAGES.get(path.suffix, None)
        if language:
            return language
        for cell_language in CellLanguage:
            if content.startswith(cell_language.file_magic_header):
                return cell_language.language
        return None

    @staticmethod
    def _adjust_path(path: Path):
        if path.suffix == ".py":
            return path
        return Path(path.as_posix() + ".py")

    def __repr__(self):
        return "NotebookLoader()"



import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient

from databricks.labs.ucx.assessment.clusters import CheckClusterMixin
from databricks.labs.ucx.framework.crawlers import CrawlerBase

logger = logging.getLogger(__name__)


@dataclass
class PipelineInfo:
    pipeline_id: str
    success: int
    failures: str
    pipeline_name: str | None = None
    creator_name: str | None = None


class PipelinesCrawler(CrawlerBase[PipelineInfo], CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema):
        super().__init__(sbe, "hive_metastore", schema, "pipelines", PipelineInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[PipelineInfo]:
        all_pipelines = list(self._ws.pipelines.list_pipelines())
        return list(self._assess_pipelines(all_pipelines))

    def _assess_pipelines(self, all_pipelines) -> Iterable[PipelineInfo]:
        for pipeline in all_pipelines:
            if not pipeline.creator_user_name:
                logger.warning(
                    f"Pipeline {pipeline.name} have Unknown creator, it means that the original creator "
                    f"has been deleted and should be re-created"
                )
            pipeline_info = PipelineInfo(
                pipeline_id=pipeline.pipeline_id,
                pipeline_name=pipeline.name,
                creator_name=pipeline.creator_user_name,
                success=1,
                failures="[]",
            )

            failures = []
            pipeline_response = self._ws.pipelines.get(pipeline.pipeline_id)
            assert pipeline_response.spec is not None
            pipeline_config = pipeline_response.spec.configuration
            if pipeline_config:
                failures.extend(self._check_spark_conf(pipeline_config, "pipeline"))
            clusters = pipeline_response.spec.clusters
            if clusters:
                self._pipeline_clusters(clusters, failures)
            pipeline_info.failures = json.dumps(failures)
            if len(failures) > 0:
                pipeline_info.success = 0
            yield pipeline_info

    def _pipeline_clusters(self, clusters, failures):
        for cluster in clusters:
            if cluster.spark_conf:
                failures.extend(self._check_spark_conf(cluster.spark_conf, "pipeline cluster"))
            # Checking if cluster config is present in cluster policies
            if cluster.policy_id:
                failures.extend(self._check_cluster_policy(cluster.policy_id, "pipeline cluster"))
            if cluster.init_scripts:
                failures.extend(self._check_cluster_init_script(cluster.init_scripts, "pipeline cluster"))

    def snapshot(self) -> Iterable[PipelineInfo]:
        return self._snapshot(self._try_fetch, self._crawl)

    def _try_fetch(self) -> Iterable[PipelineInfo]:
        for row in self._fetch(f"SELECT * FROM {self._schema}.{self._table}"):
            yield PipelineInfo(*row)



import json
import logging
import re
import uuid
from collections.abc import ValuesView
from dataclasses import dataclass
from functools import partial

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, ResourceAlreadyExists
from databricks.sdk.service.catalog import Privilege
from databricks.sdk.service.sql import EndpointConfPair

from databricks.labs.ucx.azure.resources import (
    AccessConnector,
    AzureResources,
    AzureRoleAssignment,
    PrincipalSecret,
    StorageAccount,
)
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.hive_metastore.locations import ExternalLocations

logger = logging.getLogger(__name__)


@dataclass
class StoragePermissionMapping:
    prefix: str
    client_id: str
    principal: str
    privilege: str
    type: str
    default_network_action: str = "Unknown"  # "Deny", "Allow" or "Unknown"
    # Need this directory_id/tenant_id when create UC storage credentials using service principal
    directory_id: str | None = None


class AzureResourcePermissions:
    FILENAME = 'azure_storage_account_info.csv'

    def __init__(
        self,
        installation: Installation,
        ws: WorkspaceClient,
        azurerm: AzureResources,
        external_locations: ExternalLocations,
    ):
        self._installation = installation
        self._locations = external_locations
        self._azurerm = azurerm
        self._ws = ws
        self._levels = {
            "Storage Blob Data Contributor": Privilege.WRITE_FILES,
            "Storage Blob Data Owner": Privilege.WRITE_FILES,
            "Storage Blob Data Reader": Privilege.READ_FILES,
        }
        self._permission_levels = {
            "Microsoft.Storage/storageAccounts/blobServices/containers/write": Privilege.WRITE_FILES,
            "Microsoft.Storage/storageAccounts/blobServices/containers/blobs/write": Privilege.WRITE_FILES,
            "Microsoft.Storage/storageAccounts/blobServices/containers/read": Privilege.READ_FILES,
            "Microsoft.Storage/storageAccounts/blobServices/containers/blobs/read": Privilege.READ_FILES,
        }

    def _get_permission_level(self, permission_to_match: str) -> Privilege | None:
        # String might contain '*', check for wildcard match
        pattern = re.sub(r'\*', '.*', permission_to_match)
        permission_compiled = re.compile(pattern)
        for each_level, privilege_level in self._permission_levels.items():
            # Check for storage blob permission with regex to account for star pattern
            match = permission_compiled.match(each_level)
            # If a match is found, return the privilege level, no need to check for lower levels
            if match:
                return privilege_level
        return None

    def _get_custom_role_privilege(self, role_permissions: list[str]) -> Privilege | None:
        # If both read and write privileges are found, only write privilege will be considered
        higher_privilege = None
        for each_permission in role_permissions:
            privilege = self._get_permission_level(each_permission)
            if privilege is None:
                continue
            # WRITE_FILES is the higher permission, don't need to check further
            if privilege == Privilege.WRITE_FILES:
                return privilege
            if privilege == Privilege.READ_FILES:
                higher_privilege = privilege
        return higher_privilege

    def _get_role_privilege(self, role_assignment: AzureRoleAssignment) -> Privilege | None:
        privilege = None
        # Check for custom role permissions on the storage accounts
        if role_assignment.role_permissions:
            privilege = self._get_custom_role_privilege(role_assignment.role_permissions)
        elif role_assignment.role_name in self._levels:
            privilege = self._levels[role_assignment.role_name]
        return privilege

    def _map_storage(self, storage: StorageAccount) -> ValuesView[StoragePermissionMapping]:
        logger.info(f"Fetching role assignment for {storage.name}")
        principal_spm_mapping: dict[str, StoragePermissionMapping] = {}
        for container in self._azurerm.containers(storage.id):
            for role_assignment in self._azurerm.role_assignments(str(container)):
                # Skip the role assignments that already have WRITE_FILES privilege
                spm_mapping_key = f"{container.container}_{role_assignment.principal.client_id}"
                if (
                    spm_mapping_key in principal_spm_mapping
                    and principal_spm_mapping[spm_mapping_key].privilege == Privilege.WRITE_FILES.value
                ):
                    continue
                # one principal may be assigned multiple roles with overlapping dataActions, hence appearing
                # here in duplicates. hence, role name -> permission level is not enough for the perfect scenario.
                returned_privilege = self._get_role_privilege(role_assignment)
                if not returned_privilege:
                    continue
                privilege = returned_privilege.value
                principal_spm_mapping[spm_mapping_key] = StoragePermissionMapping(
                    prefix=f"abfss://{container.container}@{container.storage_account}.dfs.core.windows.net/",
                    client_id=role_assignment.principal.client_id,
                    principal=role_assignment.principal.display_name,
                    privilege=privilege,
                    type=role_assignment.principal.type,
                    default_network_action=storage.default_network_action,
                    directory_id=role_assignment.principal.directory_id,
                )
        return principal_spm_mapping.values()

    def save_spn_permissions(self) -> str | None:
        used_storage_accounts = self._get_storage_accounts()
        if len(used_storage_accounts) == 0:
            logger.warning(
                "There are no external table present with azure storage account. "
                "Please check if assessment job is run"
            )
            return None
        storage_account_infos = []
        for storage in self._azurerm.storage_accounts():
            if storage.name not in used_storage_accounts:
                continue
            for mapping in self._map_storage(storage):
                storage_account_infos.append(mapping)
        if len(storage_account_infos) == 0:
            logger.error("No storage account found in current tenant with spn permission")
            return None
        return self._installation.save(storage_account_infos, filename=self.FILENAME)

    def _update_cluster_policy_definition(
        self,
        policy_definition: str,
        storage_accounts: list[StorageAccount],
        uber_principal: PrincipalSecret,
        inventory_database: str,
    ) -> str:
        policy_dict = json.loads(policy_definition)
        tenant_id = self._azurerm.tenant_id()
        endpoint = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"
        for storage in storage_accounts:
            policy_dict[f"spark_conf.fs.azure.account.oauth2.client.id.{storage.name}.dfs.core.windows.net"] = (
                self._policy_config(uber_principal.client.client_id)
            )
            policy_dict[f"spark_conf.fs.azure.account.oauth.provider.type.{storage.name}.dfs.core.windows.net"] = (
                self._policy_config("org.apache.hadoop.fs.azurebfs.oauth2.ClientCredsTokenProvider")
            )
            policy_dict[f"spark_conf.fs.azure.account.oauth2.client.endpoint.{storage.name}.dfs.core.windows.net"] = (
                self._policy_config(endpoint)
            )
            policy_dict[f"spark_conf.fs.azure.account.auth.type.{storage.name}.dfs.core.windows.net"] = (
                self._policy_config("OAuth")
            )
            policy_dict[f"spark_conf.fs.azure.account.oauth2.client.secret.{storage.name}.dfs.core.windows.net"] = (
                self._policy_config("{{secrets/" + inventory_database + "/uber_principal_secret}}")
            )
        return json.dumps(policy_dict)

    @staticmethod
    def _policy_config(value: str):
        return {"type": "fixed", "value": value}

    def _update_cluster_policy_with_spn(
        self,
        policy_id: str,
        storage_accounts: list[StorageAccount],
        uber_principal: PrincipalSecret,
        inventory_database: str,
    ):
        try:
            policy_definition = ""
            cluster_policy = self._ws.cluster_policies.get(policy_id)

            self._installation.save(cluster_policy, filename="policy-backup.json")

            if cluster_policy.definition is not None:
                policy_definition = self._update_cluster_policy_definition(
                    cluster_policy.definition, storage_accounts, uber_principal, inventory_database
                )
            if cluster_policy.name is not None:
                self._ws.cluster_policies.edit(policy_id, cluster_policy.name, definition=policy_definition)
        except NotFound:
            msg = f"cluster policy {policy_id} not found, please run UCX installation to create UCX cluster policy"
            raise NotFound(msg) from None

    def _update_sql_dac_with_spn(
        self,
        storage_account_info: list[StorageAccount],
        uber_principal: PrincipalSecret,
        inventory_database: str,
    ):

        warehouse_config = self._ws.warehouses.get_workspace_warehouse_config()
        sql_dac = warehouse_config.data_access_config
        if sql_dac is None:
            sql_dac = []
        tenant_id = self._azurerm.tenant_id()
        endpoint = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"
        for storage in storage_account_info:
            sql_dac.extend(
                [
                    EndpointConfPair(
                        f"spark_conf.fs.azure.account.oauth2.client.id.{storage.name}.dfs.core.windows.net",
                        uber_principal.client.client_id,
                    ),
                    EndpointConfPair(
                        f"spark_conf.fs.azure.account.oauth.provider.type.{storage.name}.dfs.core.windows.net",
                        "org.apache.hadoop.fs.azurebfs.oauth2.ClientCredsTokenProvider",
                    ),
                    EndpointConfPair(
                        f"spark_conf.fs.azure.account.oauth2.client.endpoint.{storage.name}.dfs.core.windows.net",
                        endpoint,
                    ),
                    EndpointConfPair(
                        f"spark_conf.fs.azure.account.auth.type.{storage.name}.dfs.core.windows.net", "OAuth"
                    ),
                    EndpointConfPair(
                        f"spark_conf.fs.azure.account.oauth2.client.secret.{storage.name}.dfs.core.windows.net",
                        "{{secrets/" + inventory_database + "/uber_principal_secret}}",
                    ),
                ]
            )
        self._ws.warehouses.set_workspace_warehouse_config(
            data_access_config=sql_dac,
            sql_configuration_parameters=warehouse_config.sql_configuration_parameters,
        )

    def create_uber_principal(self, prompts: Prompts):
        config = self._installation.load(WorkspaceConfig)
        inventory_database = config.inventory_database
        display_name = f"unity-catalog-migration-{inventory_database}-{self._ws.get_workspace_id()}"
        uber_principal_name = prompts.question(
            "Enter a name for the uber service principal to be created", default=display_name
        )
        policy_id = config.policy_id
        if policy_id is None:
            msg = "UCX cluster policy not found in config. Please run latest UCX installation to set cluster policy"
            logger.error(msg)
            raise ValueError(msg) from None
        if config.uber_spn_id is not None:
            logger.warning("Uber service principal already created for this workspace.")
            return
        used_storage_accounts = self._get_storage_accounts()
        if len(used_storage_accounts) == 0:
            logger.warning(
                "There are no external table present with azure storage account. "
                "Please check if assessment job is run"
            )
            return
        storage_account_info = []
        for storage in self._azurerm.storage_accounts():
            if storage.name in used_storage_accounts:
                storage_account_info.append(storage)
        logger.info("Creating service principal")
        uber_principal = self._azurerm.create_service_principal(uber_principal_name)
        self._create_scope(uber_principal, inventory_database)
        config.uber_spn_id = uber_principal.client.client_id
        logger.info(
            f"Created service principal of client_id {config.uber_spn_id}. " f"Applying permission on storage accounts"
        )
        try:
            self._apply_storage_permission(
                uber_principal.client.object_id, "STORAGE_BLOB_DATA_CONTRIBUTOR", *storage_account_info
            )
            self._installation.save(config)
            self._update_cluster_policy_with_spn(policy_id, storage_account_info, uber_principal, inventory_database)
            self._update_sql_dac_with_spn(storage_account_info, uber_principal, inventory_database)
        except PermissionError:
            self._azurerm.delete_service_principal(uber_principal.client.object_id)
        logger.info(f"Update UCX cluster policy {policy_id} with spn connection details for storage accounts")

    def _create_access_connector_for_storage_account(
        self, storage_account: StorageAccount, role_name: str = "STORAGE_BLOB_DATA_READER"
    ) -> tuple[AccessConnector, str]:
        access_connector = self._azurerm.create_or_update_access_connector(
            storage_account.id.subscription_id,
            storage_account.id.resource_group,
            f"ac-{storage_account.name}",
            storage_account.location,
            tags={"CreatedBy": "ucx"},
            wait_for_provisioning=True,
        )
        self._apply_storage_permission(access_connector.principal_id, role_name, storage_account)

        container = next(self._azurerm.containers(storage_account.id), None)
        if container is None:
            url = f"abfss://{storage_account.name}.dfs.core.windows.net/"
        else:
            url = f"abfss://{container.container}@{container.storage_account}.dfs.core.windows.net/"

        return access_connector, url

    def create_access_connectors_for_storage_accounts(self) -> list[tuple[AccessConnector, str]]:
        """Create access connectors for storage accounts

        Returns:
            list[AccessConnector, str] : The access connectors with a storage url to which it has access.
        """
        used_storage_accounts = self._get_storage_accounts()
        if len(used_storage_accounts) == 0:
            logger.warning(
                "There are no external table present with azure storage account. "
                "Please check if assessment job is run"
            )
            return []

        tasks = []
        for storage_account in self._azurerm.storage_accounts():
            if storage_account.name not in used_storage_accounts:
                continue
            task = partial(
                self._create_access_connector_for_storage_account,
                storage_account=storage_account,
                # Fine-grained access is configured within Databricks through unity
                role_name="STORAGE_BLOB_DATA_CONTRIBUTOR",
            )
            tasks.append(task)

        thread_name = "Creating access connectors for storage accounts"
        results, errors = Threads.gather(thread_name, tasks)
        if len(errors) > 0:
            raise ManyError(errors)
        return list(results)

    def _apply_storage_permission(
        self,
        principal_id: str,
        role_name: str,
        *storage_accounts: StorageAccount,
    ):
        for storage in storage_accounts:
            role_guid = str(uuid.uuid4())
            self._azurerm.apply_storage_permission(principal_id, storage, role_name, role_guid)
            logger.debug(f"{role_name} permission applied for spn {principal_id} to storage account {storage.name}")

    def _create_scope(self, uber_principal: PrincipalSecret, inventory_database: str):
        logger.info(f"Creating secret scope {inventory_database}.")
        try:
            self._ws.secrets.create_scope(inventory_database)
        except ResourceAlreadyExists:
            logger.warning(f"Secret scope {inventory_database} already exists, using the same")
        self._ws.secrets.put_secret(inventory_database, "uber_principal_secret", string_value=uber_principal.secret)

    def load(self):
        return self._installation.load(list[StoragePermissionMapping], filename=self.FILENAME)

    def _get_storage_accounts(self) -> list[str]:
        external_locations = self._locations.snapshot()
        storage_accounts = []
        for location in external_locations:
            if location.location.startswith("abfss://"):
                start = location.location.index("@")
                end = location.location.index(".dfs.core.windows.net")
                storage_acct = location.location[start + 1 : end]
                if storage_acct not in storage_accounts:
                    storage_accounts.append(storage_acct)
        return storage_accounts
